"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.Audience = void 0;
const client_utils_1 = require("@fluid-internal/client-utils");
const internal_1 = require("@fluidframework/core-utils/internal");
/**
 * Audience represents all clients connected to the op stream.
 */
class Audience extends client_utils_1.TypedEventEmitter {
    constructor() {
        super();
        this.members = new Map();
        // We are expecting this class to have many listeners, so we suppress noisy "MaxListenersExceededWarning" logging.
        super.setMaxListeners(0);
    }
    getSelf() {
        return this._currentClientId === undefined
            ? undefined
            : {
                clientId: this._currentClientId,
                client: this.getMember(this._currentClientId),
            };
    }
    setCurrentClientId(clientId) {
        if (this._currentClientId !== clientId) {
            const oldId = this._currentClientId;
            this._currentClientId = clientId;
            // this.getMember(clientId) could resolve to undefined in these two cases:
            // 1) Feature gates controlling ConnectionStateHandler() behavior are off
            // 2) we are loading from stashed state and audience is empty, but we remember and set prior clientId
            this.emit("selfChanged", oldId === undefined ? undefined : ({ clientId: oldId }), { clientId, client: this.getMember(clientId) });
        }
    }
    /**
     * Adds a new client to the audience
     */
    addMember(clientId, details) {
        // Given that signal delivery is unreliable process, we might observe same client being added twice
        // In such case we should see exactly same payload (IClient), and should not raise event twice!
        if (this.members.has(clientId)) {
            const client = this.members.get(clientId);
            (0, internal_1.assert)(JSON.stringify(client) === JSON.stringify(details), 0x4b2 /* new client has different payload from existing one */);
        }
        else {
            this.members.set(clientId, details);
            this.emit("addMember", clientId, details);
        }
    }
    /**
     * Removes a client from the audience. Only emits an event if a client is actually removed
     * @returns if a client was removed from the audience
     */
    removeMember(clientId) {
        const removedClient = this.members.get(clientId);
        if (removedClient === undefined) {
            return false;
        }
        else {
            this.members.delete(clientId);
            this.emit("removeMember", clientId, removedClient);
            return true;
        }
    }
    /**
     * Retrieves all the members in the audience.
     *
     * @remarks When the container is disconnected, this will keep returning the audience as it was last seen before the
     * container disconnected.
     */
    getMembers() {
        return new Map(this.members);
    }
    /**
     * Retrieves a specific member of the audience.
     *
     * @remarks When the container is disconnected, this will keep returning members from the audience as it was last seen
     * before the container disconnected.
     */
    getMember(clientId) {
        return this.members.get(clientId);
    }
}
exports.Audience = Audience;
//# sourceMappingURL=audience.js.map