/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { IDeltaManager } from "@fluidframework/container-definitions/internal";
import { ITelemetryBaseProperties } from "@fluidframework/core-interfaces";
import { IClient, ISequencedClient } from "@fluidframework/driver-definitions";
import { IAnyDriverError } from "@fluidframework/driver-definitions/internal";
import { type TelemetryEventCategory, ITelemetryLoggerExt, MonitoringContext } from "@fluidframework/telemetry-utils/internal";
import { ConnectionState } from "./connectionState.js";
import { IConnectionDetailsInternal, IConnectionStateChangeReason } from "./contracts.js";
import { IProtocolHandler } from "./protocol.js";
/**
 * Constructor parameter type for passing in dependencies needed by the ConnectionStateHandler
 */
export interface IConnectionStateHandlerInputs {
    logger: ITelemetryLoggerExt;
    mc: MonitoringContext;
    /**
     * Log to telemetry any change in state, included to Connecting
     */
    connectionStateChanged: (value: ConnectionState, oldState: ConnectionState, reason?: IConnectionStateChangeReason) => void;
    /**
     * Whether to expect the client to join in write mode on next connection
     */
    shouldClientJoinWrite: () => boolean;
    /**
     * (Optional) How long should we wait on our previous client's Leave op before transitioning to Connected again
     */
    maxClientLeaveWaitTime: number | undefined;
    /**
     * Log an issue encountered while in the Connecting state. details will be logged as a JSON string
     */
    logConnectionIssue: (eventName: string, category: TelemetryEventCategory, details?: ITelemetryBaseProperties) => void;
    /**
     * Callback to note that an old local client ID is still present in the Quorum that should have left and should now be considered invalid
     */
    clientShouldHaveLeft: (clientId: string) => void;
    /**
     * Some critical error was hit. Container should be closed and error logged.
     */
    onCriticalError: (error: unknown) => void;
}
/**
 * interface that connection state handler implements
 */
export interface IConnectionStateHandler {
    readonly connectionState: ConnectionState;
    /**
     * Pending clientID.
     * Changes whenever socket connection is established.
     * Resets to undefined when connection is lost
     */
    readonly pendingClientId: string | undefined;
    /**
     * clientId of a last established connection.
     * Does not reset on disconnect.
     * Changes only when new connection is established, client is fully caught up, and
     * there is no chance to ops from previous connection (i.e. if needed, we have waited and observed leave op from previous connection)
     */
    readonly clientId: string | undefined;
    containerSaved(): void;
    dispose(): void;
    initProtocol(protocol: IProtocolHandler): void;
    receivedConnectEvent(details: IConnectionDetailsInternal): void;
    receivedDisconnectEvent(reason: IConnectionStateChangeReason): void;
    establishingConnection(reason: IConnectionStateChangeReason): void;
    /**
     * Switches state to disconnected when we are still establishing connection during container.load(),
     * container connect() or reconnect and the container gets closed or disposed or disconnect happens.
     * @param reason - reason for cancelling the connection.
     */
    cancelEstablishingConnection(reason: IConnectionStateChangeReason): void;
}
export declare function createConnectionStateHandler(inputs: IConnectionStateHandlerInputs, deltaManager: IDeltaManager<unknown, unknown>, clientId?: string): ConnectionStateHandler | ConnectionStateCatchup;
export declare function createConnectionStateHandlerCore(connectedRaisedWhenCaughtUp: boolean, readClientsWaitForJoinSignal: boolean, inputs: IConnectionStateHandlerInputs, deltaManager: IDeltaManager<unknown, unknown>, clientId?: string): ConnectionStateCatchup | ConnectionStateHandler;
/**
 * Helper internal interface to abstract away Audience & Quorum
 */
interface IMembership {
    on(eventName: "addMember" | "removeMember", listener: (clientId: string, details: IClient | ISequencedClient) => void): any;
    getMember(clientId: string): undefined | unknown;
}
/**
 * Class that can be used as a base class for building IConnectionStateHandler adapters / pipeline.
 * It implements both ends of communication interfaces and passes data back and forward
 */
declare class ConnectionStateHandlerPassThrough implements IConnectionStateHandler, IConnectionStateHandlerInputs {
    protected readonly inputs: IConnectionStateHandlerInputs;
    protected readonly pimpl: IConnectionStateHandler;
    constructor(inputs: IConnectionStateHandlerInputs, pimplFactory: (handler: IConnectionStateHandlerInputs) => IConnectionStateHandler);
    get connectionState(): ConnectionState;
    get pendingClientId(): string | undefined;
    get clientId(): string | undefined;
    containerSaved(): void;
    dispose(): void;
    initProtocol(protocol: IProtocolHandler): void;
    receivedDisconnectEvent(reason: IConnectionStateChangeReason<IAnyDriverError>): void;
    establishingConnection(reason: IConnectionStateChangeReason): void;
    cancelEstablishingConnection(reason: IConnectionStateChangeReason): void;
    receivedConnectEvent(details: IConnectionDetailsInternal): void;
    get logger(): ITelemetryLoggerExt;
    get mc(): MonitoringContext;
    connectionStateChanged(value: ConnectionState, oldState: ConnectionState, reason?: IConnectionStateChangeReason): void;
    shouldClientJoinWrite(): boolean;
    get maxClientLeaveWaitTime(): number | undefined;
    logConnectionIssue(eventName: string, category: TelemetryEventCategory, details?: ITelemetryBaseProperties): void;
    clientShouldHaveLeft(clientId: string): void;
    onCriticalError(error: unknown): void;
}
/**
 * Implementation of IConnectionStateHandler pass-through adapter that waits for specific sequence number
 * before raising connected event
 */
export declare class ConnectionStateCatchup extends ConnectionStateHandlerPassThrough {
    private readonly deltaManager;
    private catchUpMonitor;
    constructor(inputs: IConnectionStateHandlerInputs, pimplFactory: (handler: IConnectionStateHandlerInputs) => IConnectionStateHandler, deltaManager: IDeltaManager<unknown, unknown>);
    private _connectionState;
    get connectionState(): ConnectionState;
    connectionStateChanged(value: ConnectionState, oldState: ConnectionState, reason?: IConnectionStateChangeReason<IAnyDriverError>): void;
    private readonly transitionToConnectedState;
}
/**
 * In the lifetime of a container, the connection will likely disconnect and reconnect periodically.
 * This class ensures that any ops sent by this container instance on previous connection are either
 * sequenced or blocked by the server before emitting the new "connected" event and allowing runtime to resubmit ops.
 *
 * Each connection is assigned a clientId by the service, and the connection is book-ended by a Join and a Leave op
 * generated by the service. Due to the distributed nature of the Relay Service, in the case of reconnect we cannot
 * make any assumptions about ordering of operations between the old and new connections - i.e. new Join op could
 * be sequenced before old Leave op (and some acks from pending ops that were in flight when we disconnected).
 *
 * The job of this class is to encapsulate the transition period during reconnect, which is identified by
 * ConnectionState.CatchingUp. Specifically, before moving to Connected state with the new clientId, it ensures that:
 *
 * a. We process the Leave op for the previous clientId. This allows us to properly handle any acks from in-flight ops
 * that got sequenced with the old clientId (we'll recognize them as local ops). After the Leave op, any other
 * pending ops can safely be submitted with the new clientId without fear of duplication in the sequenced op stream.
 *
 * b. We process the Join op for the new clientId (identified when the underlying connection was first established),
 * indicating the service is ready to sequence ops sent with the new clientId.
 *
 * c. We process all ops known at the time the underlying connection was established (so we are "caught up")
 *
 * For (a) we give up waiting after some time (same timeout as server uses), and go ahead and transition to Connected.
 *
 * For (b) we log telemetry if it takes too long, but still only transition to Connected when the Join op/signal is
 * processed.
 *
 * For (c) this is optional behavior, controlled by the parameters of receivedConnectEvent
 */
export declare class ConnectionStateHandler implements IConnectionStateHandler {
    private readonly handler;
    private readonly readClientsWaitForJoinSignal;
    private _connectionState;
    private _pendingClientId;
    /**
     * Tracks that we observe the "leave" op within the timeout for our previous clientId (see comment on ConnectionStateHandler class)
     * ! This ensures we do not switch to a new clientId until we process all potential messages from old clientId
     * ! i.e. We will always see the "leave" op for a client after we have seen all the ops it has sent
     * ! This check helps prevent the same op from being resubmitted by the PendingStateManager upon reconnecting
     */
    private readonly prevClientLeftTimer;
    /**
     * Tracks that we observe our own "join" op within the timeout after receiving a "connected" event from the DeltaManager
     */
    private readonly joinTimer;
    private protocol?;
    private connection?;
    private _clientId?;
    /**
     * Track how long we waited to see "leave" op for previous clientId
     */
    private waitEvent;
    get connectionState(): ConnectionState;
    get clientId(): string | undefined;
    get pendingClientId(): string | undefined;
    constructor(handler: IConnectionStateHandlerInputs, readClientsWaitForJoinSignal: boolean, clientIdFromPausedSession?: string);
    private startjoinTimer;
    private stopjoinTimer;
    private get waitingForLeaveOp();
    dispose(): void;
    containerSaved(): void;
    private receivedAddMemberEvent;
    private applyForConnectedState;
    private receivedRemoveMemberEvent;
    receivedDisconnectEvent(reason: IConnectionStateChangeReason<IAnyDriverError>): void;
    cancelEstablishingConnection(reason: IConnectionStateChangeReason): void;
    establishingConnection(reason: IConnectionStateChangeReason): void;
    /**
     * Tells if need to wait for "self" to show up in audience.
     * @returns - true if we should wait for "self" to appear in audience.
     * false is returned only for "read" connections, and only if this.readClientsWaitForJoinSignal is false.
     */
    private shouldWaitForSelf;
    /**
     * The "connect" event indicates the connection to the Relay Service is live.
     * However, some additional conditions must be met before we can fully transition to
     * "Connected" state. This function handles that interim period, known as "Connecting" state.
     * @param details - Connection details returned from the Relay Service
     * @param deltaManager - DeltaManager to be used for delaying Connected transition until caught up.
     * If it's undefined, then don't delay and transition to Connected as soon as Leave/Join op are accounted for
     */
    receivedConnectEvent(details: IConnectionDetailsInternal): void;
    private setConnectionState;
    protected get membership(): IMembership | undefined;
    initProtocol(protocol: IProtocolHandler): void;
    protected hasMember(clientId?: string): boolean;
}
export {};
//# sourceMappingURL=connectionStateHandler.d.ts.map