/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { AttachState, IAudience, ICriticalContainerError } from "@fluidframework/container-definitions";
import type { ICodeDetailsLoader, IContainer, IContainerEvents, IContainerLoadMode, IDeltaManager, IFluidCodeDetails, ReadOnlyInfo, ILoaderOptions } from "@fluidframework/container-definitions/internal";
import { FluidObject, IRequest, ITelemetryBaseProperties } from "@fluidframework/core-interfaces";
import { IClientDetails, IQuorumClients } from "@fluidframework/driver-definitions";
import { IDocumentServiceFactory, IResolvedUrl, IUrlResolver, IDocumentMessage, ISequencedDocumentMessage } from "@fluidframework/driver-definitions/internal";
import { ITelemetryLoggerExt, EventEmitterWithErrorHandling } from "@fluidframework/telemetry-utils/internal";
import { ConnectionState } from "./connectionState.js";
import { ProtocolHandlerBuilder } from "./protocol.js";
import { type IPendingContainerState } from "./serializedStateManager.js";
/**
 * @internal
 */
export interface IContainerLoadProps {
    /**
     * The resolved url of the container being loaded
     */
    readonly resolvedUrl: IResolvedUrl;
    /**
     * Control which snapshot version to load from.  See IParsedUrl for detailed information.
     */
    readonly version: string | undefined;
    /**
     * Loads the Container in paused state if true, unpaused otherwise.
     */
    readonly loadMode?: IContainerLoadMode;
    /**
     * The pending state serialized from a previous container instance
     */
    readonly pendingLocalState?: IPendingContainerState;
}
/**
 * @internal
 */
export interface IContainerCreateProps {
    /**
     * Disables the Container from reconnecting if false, allows reconnect otherwise.
     */
    readonly canReconnect?: boolean;
    /**
     * Client details provided in the override will be merged over the default client.
     */
    readonly clientDetailsOverride?: IClientDetails;
    /**
     * The url resolver used by the loader for resolving external urls
     * into Fluid urls such that the container specified by the
     * external url can be loaded.
     */
    readonly urlResolver: IUrlResolver;
    /**
     * The document service factory take the Fluid url provided
     * by the resolved url and constructs all the necessary services
     * for communication with the container's server.
     */
    readonly documentServiceFactory: IDocumentServiceFactory;
    /**
     * The code loader handles loading the necessary code
     * for running a container once it is loaded.
     */
    readonly codeLoader: ICodeDetailsLoader;
    /**
     * A property bag of options used by various layers
     * to control features
     */
    readonly options: ILoaderOptions;
    /**
     * Scope is provided to all container and is a set of shared
     * services for container's to integrate with their host environment.
     */
    readonly scope: FluidObject;
    /**
     * The logger downstream consumers should construct their loggers from
     */
    readonly subLogger: ITelemetryLoggerExt;
    /**
     * Optional property for allowing the container to use a custom
     * protocol implementation for handling the quorum and/or the audience.
     */
    readonly protocolHandlerBuilder?: ProtocolHandlerBuilder;
}
/**
 * Waits until container connects to delta storage and gets up-to-date.
 *
 * Useful when resolving URIs and hitting 404, due to container being loaded from (stale) snapshot and not being
 * up to date. Host may chose to wait in such case and retry resolving URI.
 *
 * Warning: Will wait infinitely for connection to establish if there is no connection.
 * May result in deadlock if Container.disconnect() is called and never followed by a call to Container.connect().
 *
 * @returns `true`: container is up to date, it processed all the ops that were know at the time of first connection.
 *
 * `false`: storage does not provide indication of how far the client is. Container processed all the ops known to it,
 * but it maybe still behind.
 *
 * @throws an error beginning with `"Container closed"` if the container is closed before it catches up.
 * @legacy
 * @alpha
 */
export declare function waitContainerToCatchUp(container: IContainer): Promise<boolean>;
/**
 * Helper function to report to telemetry cases where operation takes longer than expected (200ms)
 * @param logger - logger to use
 * @param eventName - event name
 * @param action - functor to call and measure
 */
export declare function ReportIfTooLong(logger: ITelemetryLoggerExt, eventName: string, action: () => Promise<ITelemetryBaseProperties>): Promise<void>;
export declare class Container extends EventEmitterWithErrorHandling<IContainerEvents> implements IContainer, IContainerExperimental {
    /**
     * Load an existing container.
     */
    static load(loadProps: IContainerLoadProps, createProps: IContainerCreateProps): Promise<Container>;
    /**
     * Create a new container in a detached state.
     */
    static createDetached(createProps: IContainerCreateProps, codeDetails: IFluidCodeDetails): Promise<Container>;
    /**
     * Create a new container in a detached state that is initialized with a
     * snapshot from a previous detached container.
     * @param createProps - Config options for this new container instance
     * @param snapshot - A stringified {@link IPendingDetachedContainerState}, e.g. generated via {@link serialize}
     */
    static rehydrateDetachedFromSnapshot(createProps: IContainerCreateProps, snapshot: string): Promise<Container>;
    private readonly _canReconnect;
    private readonly clientDetailsOverride;
    private readonly urlResolver;
    private readonly serviceFactory;
    private readonly codeLoader;
    private readonly options;
    private readonly scope;
    private readonly subLogger;
    private readonly detachedBlobStorage;
    private readonly protocolHandlerBuilder;
    private readonly client;
    private readonly mc;
    /**
     * Used by the RelativeLoader to spawn a new Container for the same document.  Used to create the summarizing client.
     */
    readonly clone: (loadProps: IContainerLoadProps, createParamOverrides: Partial<IContainerCreateProps>) => Promise<Container>;
    /**
     * Lifecycle state of the container, used mainly to prevent re-entrancy and telemetry
     *
     * States are allowed to progress to further states:
     * "loading" - "loaded" - "closing" - "disposing" - "closed" - "disposed"
     *
     * For example, moving from "closed" to "disposing" is not allowed since it is an earlier state.
     *
     * loading: Container has been created, but is not yet in normal/loaded state
     * loaded: Container is in normal/loaded state
     * closing: Container has started closing process (for re-entrancy prevention)
     * disposing: Container has started disposing process (for re-entrancy prevention)
     * closed: Container has closed
     * disposed: Container has been disposed
     */
    private _lifecycleState;
    private setLoaded;
    get closed(): boolean;
    protected get loaded(): boolean;
    get disposed(): boolean;
    private readonly storageAdapter;
    private readonly _deltaManager;
    private service;
    private _runtime;
    private get runtime();
    private _protocolHandler;
    private get protocolHandler();
    /**
     * During initialization we pause the inbound queues. We track this state to ensure we only call resume once
     */
    private inboundQueuePausedFromInit;
    private connectionCount;
    private readonly connectionTransitionTimes;
    private _loadedFromVersion;
    private _dirtyContainer;
    private attachmentData;
    private readonly serializedStateManager;
    private readonly _containerId;
    private lastVisible;
    private readonly visibilityEventHandler;
    private readonly connectionStateHandler;
    private readonly clientsWhoShouldHaveLeft;
    private _containerMetadata;
    private setAutoReconnectTime;
    private noopHeuristic;
    private get connectionMode();
    get resolvedUrl(): IResolvedUrl | undefined;
    get readOnlyInfo(): ReadOnlyInfo;
    get containerMetadata(): Record<string, string>;
    /**
     * Sends signal to runtime (and data stores) to be read-only.
     * Hosts may have read only views, indicating to data stores that no edits are allowed.
     * This is independent from this._readonlyPermissions (permissions) and this.connectionMode
     * (server can return "write" mode even when asked for "read")
     * Leveraging same "readonly" event as runtime & data stores should behave the same in such case
     * as in read-only permissions.
     * But this.active can be used by some DDSes to figure out if ops can be sent
     * (for example, read-only view still participates in code proposals / upgrades decisions)
     *
     * Forcing Readonly does not prevent DDS from generating ops. It is up to user code to honour
     * the readonly flag. If ops are generated, they will accumulate locally and not be sent. If
     * there are pending in the outbound queue, it will stop sending until force readonly is
     * cleared.
     *
     * @param readonly - set or clear force readonly.
     */
    forceReadonly(readonly: boolean): void;
    get deltaManager(): IDeltaManager<ISequencedDocumentMessage, IDocumentMessage>;
    get connectionState(): ConnectionState;
    private get connected();
    /**
     * clientId of the latest connection. Changes only once client is connected, caught up and fully loaded.
     * Changes to clientId are delayed through container loading sequence and delived once container is fully loaded.
     * clientId does not reset on lost connection - old value persists until new connection is fully established.
     */
    get clientId(): string | undefined;
    private get isInteractiveClient();
    private supportGetSnapshotApi;
    /**
     * Get the code details that are currently specified for the container.
     * @returns The current code details if any are specified, undefined if none are specified.
     */
    getSpecifiedCodeDetails(): IFluidCodeDetails | undefined;
    private _loadedCodeDetails;
    /**
     * Get the code details that were used to load the container.
     * @returns The code details that were used to load the container if it is loaded, undefined if it is not yet
     * loaded.
     */
    getLoadedCodeDetails(): IFluidCodeDetails | undefined;
    private _loadedModule;
    /**
     * Retrieves the audience associated with the document
     */
    get audience(): IAudience;
    /**
     * Returns true if container is dirty.
     * Which means data loss if container is closed at that same moment
     * Most likely that happens when there is no network connection to Relay Service
     */
    get isDirty(): boolean;
    /**
     * {@inheritDoc @fluidframework/container-definitions#IContainer.entryPoint}
     */
    getEntryPoint(): Promise<FluidObject>;
    private readonly _lifecycleEvents;
    constructor(createProps: IContainerCreateProps, loadProps?: Pick<IContainerLoadProps, "pendingLocalState">);
    /**
     * Retrieves the quorum associated with the document
     */
    getQuorum(): IQuorumClients;
    dispose(error?: ICriticalContainerError): void;
    close(error?: ICriticalContainerError): void;
    private verifyClosedAfterCalls;
    private verifyClosedAfter;
    private closeCore;
    private _disposed;
    private disposeCore;
    closeAndGetPendingLocalState(stopBlobAttachingSignal?: AbortSignal): Promise<string>;
    /**
     * Serialize current container state required to rehydrate to the same position without dataloss.
     * Note: The container must already be attached. For detached containers use {@link serialize}
     * @returns stringified {@link IPendingContainerState} for the container
     */
    getPendingLocalState(): Promise<string>;
    private getPendingLocalStateCore;
    get attachState(): AttachState;
    /**
     * Serialize current container state required to rehydrate to the same position without dataloss.
     * Note: The container must be detached and not closed. For attached containers use
     * {@link getPendingLocalState} or {@link closeAndGetPendingLocalState}
     * @returns stringified {@link IPendingDetachedContainerState} for the container
     */
    serialize(): string;
    readonly attach: (request: IRequest, attachProps?: {
        deltaConnection?: "none" | "delayed" | undefined;
    } | undefined) => Promise<void>;
    private setAutoReconnectInternal;
    connect(): void;
    private connectInternal;
    disconnect(): void;
    private disconnectInternal;
    private resumeInternal;
    readonly getAbsoluteUrl: (relativeUrl: string) => Promise<string | undefined>;
    proposeCodeDetails(codeDetails: IFluidCodeDetails): Promise<boolean>;
    private processCodeProposal;
    /**
     * Determines if the currently loaded module satisfies the incoming constraint code details
     */
    private satisfies;
    private connectToDeltaStream;
    private readonly metadataUpdateHandler;
    private createDocumentService;
    /**
     * Load container.
     *
     * @param specifiedVersion - Version SHA to load snapshot. If not specified, will fetch the latest snapshot.
     */
    private load;
    private createDetached;
    private rehydrateDetachedFromSnapshot;
    private initializeProtocolStateFromSnapshot;
    private initializeProtocolState;
    private captureProtocolSummary;
    private getCodeDetailsFromQuorum;
    private static setupClient;
    /**
     * Returns true if connection is active, i.e. it's "write" connection and
     * container runtime was notified about this connection (i.e. we are up-to-date and could send ops).
     * This happens after client received its own joinOp and thus is in the quorum.
     * If it's not true, runtime is not in position to send ops.
     */
    private activeConnection;
    private createDeltaManager;
    private attachDeltaManagerOpHandler;
    private logConnectionStateChangeTelemetry;
    private propagateConnectionState;
    private submitContainerMessage;
    /**
     * Gets the `clientSequenceNumber` of last message in a batch.
     */
    private submitBatch;
    private submitSummaryMessage;
    private submitMessage;
    /**
     * Processes incoming delta messages
     * @param message - delta message received from the server
     */
    private processRemoteMessage;
    private submitSignal;
    private processSignal;
    private instantiateRuntime;
    private readonly updateDirtyContainerState;
    /**
     * Set the connected state of the ContainerContext
     * This controls the "connected" state of the ContainerRuntime as well
     * @param connected - Is the container currently connected?
     * @param readonly - Is the container in readonly mode?
     */
    private setContextConnectedState;
    private handleDeltaConnectionArg;
}
/**
 * IContainer interface that includes experimental features still under development.
 * @internal
 */
export interface IContainerExperimental extends IContainer {
    /**
     * Get pending state from container. WARNING: misuse of this API can result in duplicate op
     * submission and potential document corruption. The blob returned MUST be deleted if and when this
     * container emits a "connected" event.
     * @returns serialized blob that can be passed to Loader.resolve()
     */
    getPendingLocalState?(): Promise<string>;
    /**
     * Closes the container and returns serialized local state intended to be
     * given to a newly loaded container.
     */
    closeAndGetPendingLocalState?(stopBlobAttachingSignal?: AbortSignal): Promise<string>;
}
//# sourceMappingURL=container.d.ts.map