"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DebugLogger = void 0;
const client_utils_1 = require("@fluid-internal/client-utils");
const internal_1 = require("@fluidframework/telemetry-utils/internal");
// This import style is necessary to ensure the emitted JS code works in both CJS and ESM.
const debug_1 = __importDefault(require("debug"));
const { debug: registerDebug } = debug_1.default;
/**
 * Implementation of debug logger
 */
class DebugLogger {
    /**
     * Mix in debug logger with another logger.
     * Returned logger will output events to both newly created debug logger, as well as base logger
     * @param namespace - Telemetry event name prefix to add to all events
     * @param properties - Base properties to add to all events
     * @param propertyGetters - Getters to add additional properties to all events
     * @param baseLogger - Base logger to output events (in addition to debug logger being created). Can be undefined.
     */
    static mixinDebugLogger(namespace, baseLogger, properties) {
        // Setup base logger upfront, such that host can disable it (if needed)
        const debug = registerDebug(namespace);
        // Create one for errors that is always enabled
        // It can be silenced by replacing console.error if the debug namespace is not enabled.
        const debugErr = registerDebug(namespace);
        debugErr.log = function (...args) {
            if (debug.enabled === true) {
                // if the namespace is enabled, just use the default logger
                registerDebug.log(...args);
            }
            else {
                // other wise, use the console logger (which could be replaced and silenced)
                console.error(...args);
            }
        };
        debugErr.enabled = true;
        return (0, internal_1.createMultiSinkLogger)({
            namespace,
            loggers: [baseLogger, new DebugLogger(debug, debugErr)],
            properties,
            tryInheritProperties: true,
        });
    }
    constructor(debug, debugErr) {
        this.debug = debug;
        this.debugErr = debugErr;
    }
    /**
     * Send an event to debug loggers
     *
     * @param event - the event to send
     */
    send(event) {
        const newEvent = { ...event };
        const isError = newEvent.category === "error";
        let logger = isError ? this.debugErr : this.debug;
        // Use debug's coloring schema for base of the event
        const index = event.eventName.lastIndexOf(internal_1.eventNamespaceSeparator);
        const name = event.eventName.slice(Math.max(0, index + 1));
        if (index > 0) {
            logger = logger.extend(event.eventName.slice(0, index));
        }
        newEvent.eventName = undefined;
        let tick = "";
        tick = `tick=${(0, internal_1.formatTick)((0, client_utils_1.performanceNow)())}`;
        // Extract stack to put it last, but also to avoid escaping '\n' in it by JSON.stringify below
        const stack = newEvent.stack ?? "";
        newEvent.stack = undefined;
        // Watch out for circular references - they can come from two sources
        // 1) error object - we do not control it and should remove it and retry
        // 2) properties supplied by telemetry caller - that's a bug that should be addressed!
        let payload;
        try {
            payload = JSON.stringify(newEvent);
        }
        catch {
            newEvent.error = undefined;
            payload = JSON.stringify(newEvent);
        }
        if (payload === "{}") {
            payload = "";
        }
        // Force errors out, to help with diagnostics
        if (isError) {
            logger.enabled = true;
        }
        // Print multi-line.
        // eslint-disable-next-line @typescript-eslint/no-base-to-string
        logger(`${name} ${payload} ${tick} ${stack}`);
    }
}
exports.DebugLogger = DebugLogger;
//# sourceMappingURL=debugLogger.js.map