/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { IContainer, IFluidCodeDetails, IFluidModule, IHostLoader, ILoader, ILoaderOptions, IProvideFluidCodeDetailsComparer } from "@fluidframework/container-definitions/internal";
import { FluidObject, IConfigProviderBase, IRequest, ITelemetryBaseLogger } from "@fluidframework/core-interfaces";
import { IClientDetails } from "@fluidframework/driver-definitions";
import { IDocumentServiceFactory, IUrlResolver } from "@fluidframework/driver-definitions/internal";
import { ITelemetryLoggerExt } from "@fluidframework/telemetry-utils/internal";
import { Container } from "./container.js";
import { ProtocolHandlerBuilder } from "./protocol.js";
/**
 * @internal
 */
export declare class RelativeLoader implements ILoader {
    private readonly container;
    private readonly loader;
    constructor(container: Container, loader: ILoader | undefined);
    resolve(request: IRequest): Promise<IContainer>;
}
/**
 * @deprecated IFluidModuleWithDetails interface is moved to
 * {@link @fluidframework/container-definitions#IFluidModuleWithDetails}
 * to have all the code loading modules in one package. #8193
 * Encapsulates a module entry point with corresponding code details.
 * @legacy
 * @alpha
 */
export interface IFluidModuleWithDetails {
    /**
     * Fluid code module that implements the runtime factory needed to instantiate the container runtime.
     */
    module: IFluidModule;
    /**
     * Code details associated with the module. Represents a document schema this module supports.
     * If the code loader implements the {@link @fluidframework/core-interfaces#IFluidCodeDetailsComparer} interface,
     * it'll be called to determine whether the module code details satisfy the new code proposal in the quorum.
     */
    details: IFluidCodeDetails;
}
/**
 * @deprecated ICodeDetailsLoader interface is moved to {@link @fluidframework/container-definitions#ICodeDetailsLoader}
 * to have code loading modules in one package. #8193
 * Fluid code loader resolves a code module matching the document schema, i.e. code details, such as
 * a package name and package version range.
 * @legacy
 * @alpha
 */
export interface ICodeDetailsLoader extends Partial<IProvideFluidCodeDetailsComparer> {
    /**
     * Load the code module (package) that is capable to interact with the document.
     *
     * @param source - Code proposal that articulates the current schema the document is written in.
     * @returns Code module entry point along with the code details associated with it.
     */
    load(source: IFluidCodeDetails): Promise<IFluidModuleWithDetails>;
}
/**
 * Services and properties necessary for creating a loader
 * @legacy
 * @alpha
 */
export interface ILoaderProps {
    /**
     * The url resolver used by the loader for resolving external urls
     * into Fluid urls such that the container specified by the
     * external url can be loaded.
     */
    readonly urlResolver: IUrlResolver;
    /**
     * The document service factory take the Fluid url provided
     * by the resolved url and constructs all the necessary services
     * for communication with the container's server.
     */
    readonly documentServiceFactory: IDocumentServiceFactory;
    /**
     * The code loader handles loading the necessary code
     * for running a container once it is loaded.
     */
    readonly codeLoader: ICodeDetailsLoader;
    /**
     * A property bag of options used by various layers
     * to control features
     */
    readonly options?: ILoaderOptions;
    /**
     * Scope is provided to all container and is a set of shared
     * services for container's to integrate with their host environment.
     */
    readonly scope?: FluidObject;
    /**
     * The logger that all telemetry should be pushed to.
     */
    readonly logger?: ITelemetryBaseLogger;
    /**
     * The configuration provider which may be used to control features.
     */
    readonly configProvider?: IConfigProviderBase;
    /**
     * Optional property for allowing the container to use a custom
     * protocol implementation for handling the quorum and/or the audience.
     */
    readonly protocolHandlerBuilder?: ProtocolHandlerBuilder;
}
/**
 * Services and properties used by and exposed by the loader
 * @legacy
 * @alpha
 */
export interface ILoaderServices {
    /**
     * The url resolver used by the loader for resolving external urls
     * into Fluid urls such that the container specified by the
     * external url can be loaded.
     */
    readonly urlResolver: IUrlResolver;
    /**
     * The document service factory take the Fluid url provided
     * by the resolved url and constructs all the necessary services
     * for communication with the container's server.
     */
    readonly documentServiceFactory: IDocumentServiceFactory;
    /**
     * The code loader handles loading the necessary code
     * for running a container once it is loaded.
     */
    readonly codeLoader: ICodeDetailsLoader;
    /**
     * A property bag of options used by various layers
     * to control features
     */
    readonly options: ILoaderOptions;
    /**
     * Scope is provided to all container and is a set of shared
     * services for container's to integrate with their host environment.
     */
    readonly scope: FluidObject;
    /**
     * The logger downstream consumers should construct their loggers from
     */
    readonly subLogger: ITelemetryLoggerExt;
    /**
     * Optional property for allowing the container to use a custom
     * protocol implementation for handling the quorum and/or the audience.
     */
    readonly protocolHandlerBuilder?: ProtocolHandlerBuilder;
}
/**
 * Manages Fluid resource loading
 * @legacy
 * @alpha
 *
 * @remarks The Loader class is deprecated and will be removed in a future release. Use the free-form functions instead (See issue #24450 for more details).
 */
export declare class Loader implements IHostLoader {
    readonly services: ILoaderServices;
    private readonly mc;
    constructor(loaderProps: ILoaderProps);
    createDetachedContainer(codeDetails: IFluidCodeDetails, createDetachedProps?: {
        canReconnect?: boolean;
        clientDetailsOverride?: IClientDetails;
    }): Promise<IContainer>;
    rehydrateDetachedContainerFromSnapshot(snapshot: string, createDetachedProps?: {
        canReconnect?: boolean;
        clientDetailsOverride?: IClientDetails;
    }): Promise<IContainer>;
    resolve(request: IRequest, pendingLocalState?: string): Promise<IContainer>;
    private resolveCore;
    private loadContainer;
}
//# sourceMappingURL=loader.d.ts.map