"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.Loader = exports.RelativeLoader = void 0;
const internal_1 = require("@fluidframework/container-definitions/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const uuid_1 = require("uuid");
const container_js_1 = require("./container.js");
const debugLogger_js_1 = require("./debugLogger.js");
const packageVersion_js_1 = require("./packageVersion.js");
const utils_js_1 = require("./utils.js");
function ensureResolvedUrlDefined(resolved) {
    if (resolved === undefined) {
        throw new Error(`Object is not a IResolveUrl.`);
    }
}
/**
 * @internal
 */
class RelativeLoader {
    constructor(container, loader) {
        this.container = container;
        this.loader = loader;
    }
    async resolve(request) {
        if (request.url.startsWith("/")) {
            ensureResolvedUrlDefined(this.container.resolvedUrl);
            const container = await this.container.clone({
                resolvedUrl: { ...this.container.resolvedUrl },
                // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
                version: request.headers?.[internal_1.LoaderHeader.version] ?? undefined,
                // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
                loadMode: request.headers?.[internal_1.LoaderHeader.loadMode],
            }, {
                // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
                canReconnect: request.headers?.[internal_1.LoaderHeader.reconnect],
                // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
                clientDetailsOverride: request.headers?.[internal_1.LoaderHeader.clientDetails],
            });
            return container;
        }
        if (this.loader === undefined) {
            throw new Error("Cannot resolve external containers");
        }
        return this.loader.resolve(request);
    }
}
exports.RelativeLoader = RelativeLoader;
/**
 * Manages Fluid resource loading
 * @legacy
 * @alpha
 *
 * @remarks The Loader class is deprecated and will be removed in a future release. Use the free-form functions instead (See issue #24450 for more details).
 */
class Loader {
    constructor(loaderProps) {
        const { urlResolver, documentServiceFactory, codeLoader, options, scope, logger, configProvider, protocolHandlerBuilder, } = loaderProps;
        const telemetryProps = {
            loaderId: (0, uuid_1.v4)(),
            loaderVersion: packageVersion_js_1.pkgVersion,
        };
        const subMc = (0, internal_2.mixinMonitoringContext)(debugLogger_js_1.DebugLogger.mixinDebugLogger("fluid:telemetry", logger, {
            all: telemetryProps,
        }), internal_2.sessionStorageConfigProvider.value, configProvider);
        this.services = {
            urlResolver,
            documentServiceFactory,
            codeLoader,
            options: options ?? {},
            scope: options?.provideScopeLoader === false ? { ...scope } : { ...scope, ILoader: this },
            protocolHandlerBuilder,
            subLogger: subMc.logger,
        };
        this.mc = (0, internal_2.createChildMonitoringContext)({
            logger: this.services.subLogger,
            namespace: "Loader",
        });
    }
    async createDetachedContainer(codeDetails, createDetachedProps) {
        return container_js_1.Container.createDetached({
            ...createDetachedProps,
            ...this.services,
        }, codeDetails);
    }
    async rehydrateDetachedContainerFromSnapshot(snapshot, createDetachedProps) {
        return container_js_1.Container.rehydrateDetachedFromSnapshot({
            ...createDetachedProps,
            ...this.services,
        }, snapshot);
    }
    async resolve(request, pendingLocalState) {
        const eventName = pendingLocalState === undefined ? "Resolve" : "ResolveWithPendingState";
        return internal_2.PerformanceEvent.timedExecAsync(this.mc.logger, { eventName }, async () => {
            return this.resolveCore(request, (0, utils_js_1.getAttachedContainerStateFromSerializedContainer)(pendingLocalState));
        });
    }
    async resolveCore(request, pendingLocalState) {
        const resolvedAsFluid = await this.services.urlResolver.resolve(request);
        ensureResolvedUrlDefined(resolvedAsFluid);
        // Parse URL into data stores
        const parsed = (0, utils_js_1.tryParseCompatibleResolvedUrl)(resolvedAsFluid.url);
        if (parsed === undefined) {
            throw new Error(`Invalid URL ${resolvedAsFluid.url}`);
        }
        if (pendingLocalState !== undefined) {
            const parsedPendingUrl = (0, utils_js_1.tryParseCompatibleResolvedUrl)(pendingLocalState.url);
            if (parsedPendingUrl?.id !== parsed.id ||
                parsedPendingUrl?.path.replace(/\/$/, "") !== parsed.path.replace(/\/$/, "")) {
                const message = `URL ${resolvedAsFluid.url} does not match pending state URL ${pendingLocalState.url}`;
                throw new Error(message);
            }
        }
        request.headers ??= {};
        // If set in both query string and headers, use query string.  Also write the value from the query string into the header either way.
        // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
        request.headers[internal_1.LoaderHeader.version] =
            parsed.version ?? request.headers[internal_1.LoaderHeader.version];
        return this.loadContainer(request, resolvedAsFluid, pendingLocalState);
    }
    async loadContainer(request, resolvedUrl, pendingLocalState) {
        return container_js_1.Container.load({
            resolvedUrl,
            // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
            version: request.headers?.[internal_1.LoaderHeader.version] ?? undefined,
            // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
            loadMode: request.headers?.[internal_1.LoaderHeader.loadMode],
            pendingLocalState,
        }, {
            // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
            canReconnect: request.headers?.[internal_1.LoaderHeader.reconnect],
            // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
            clientDetailsOverride: request.headers?.[internal_1.LoaderHeader.clientDetails],
            ...this.services,
        });
    }
}
exports.Loader = Loader;
//# sourceMappingURL=loader.js.map