"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.validateRuntimeCompatibility = exports.runtimeSupportRequirements = exports.loaderCompatDetailsForRuntime = exports.loaderCoreCompatDetails = void 0;
const client_utils_1 = require("@fluid-internal/client-utils");
const internal_1 = require("@fluidframework/telemetry-utils/internal");
const packageVersion_js_1 = require("./packageVersion.js");
/**
 * The core compatibility details of the Loader layer that is the same across all layer boundaries.
 * @internal
 */
exports.loaderCoreCompatDetails = {
    /**
     * The package version of the Loader layer.
     */
    pkgVersion: packageVersion_js_1.pkgVersion,
    /**
     * The current generation of the Loader layer.
     */
    generation: 1,
};
/**
 * Loader's compatibility details that is exposed to the Runtime layer.
 * @internal
 */
exports.loaderCompatDetailsForRuntime = {
    ...exports.loaderCoreCompatDetails,
    /**
     * The features supported by the Loader layer across the Loader / Runtime boundary.
     */
    supportedFeatures: new Set(),
};
/**
 * The requirements that the Runtime layer must meet to be compatible with this Loader.
 * @internal
 */
exports.runtimeSupportRequirements = {
    /**
     * Minimum generation that Runtime must be at to be compatible with Loader. Note that 0 is used here for
     * Runtime layers before the introduction of the layer compatibility enforcement.
     */
    minSupportedGeneration: 0,
    /**
     * The features that the Runtime must support to be compatible with Loader.
     */
    requiredFeatures: [],
};
/**
 * Validates that the Runtime layer is compatible with the Loader.
 * @internal
 */
function validateRuntimeCompatibility(maybeRuntimeCompatDetails, disposeFn) {
    const layerCheckResult = (0, client_utils_1.checkLayerCompatibility)(exports.runtimeSupportRequirements, maybeRuntimeCompatDetails);
    if (!layerCheckResult.isCompatible) {
        const error = new internal_1.UsageError("Loader is not compatible with Runtime", {
            errorDetails: JSON.stringify({
                loaderVersion: exports.loaderCompatDetailsForRuntime.pkgVersion,
                runtimeVersion: maybeRuntimeCompatDetails?.pkgVersion,
                loaderGeneration: exports.loaderCompatDetailsForRuntime.generation,
                runtimeGeneration: maybeRuntimeCompatDetails?.generation,
                minSupportedGeneration: exports.runtimeSupportRequirements.minSupportedGeneration,
                isGenerationCompatible: layerCheckResult.isGenerationCompatible,
                unsupportedFeatures: layerCheckResult.unsupportedFeatures,
            }),
        });
        disposeFn(error);
        throw error;
    }
}
exports.validateRuntimeCompatibility = validateRuntimeCompatibility;
//# sourceMappingURL=loaderLayerCompatState.js.map