/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { TypedEventEmitter } from "@fluid-internal/client-utils";
import { IQuorumClients, ISequencedClient } from "@fluidframework/driver-definitions";
import { ISequencedDocumentMessage, ICommittedProposal, IQuorum, IQuorumProposals, ISequencedProposal } from "@fluidframework/driver-definitions/internal";
/**
 * Snapshot format for a QuorumClients
 * @legacy
 * @alpha
 */
export type QuorumClientsSnapshot = [string, ISequencedClient][];
/**
 * Snapshot format for a QuorumProposals
 * @legacy
 * @alpha
 */
export type QuorumProposalsSnapshot = {
    proposals: [number, ISequencedProposal, string[]][];
    values: [string, ICommittedProposal][];
};
/**
 * Snapshot format for a Quorum
 * @legacy
 * @alpha
 */
export interface IQuorumSnapshot {
    members: QuorumClientsSnapshot;
    proposals: QuorumProposalsSnapshot["proposals"];
    values: QuorumProposalsSnapshot["values"];
}
/**
 * The QuorumClients is used to track members joining and leaving the collaboration session.
 * @internal
 */
export declare class QuorumClients extends TypedEventEmitter<IQuorumClients["on"]> implements IQuorumClients {
    private readonly members;
    private isDisposed;
    get disposed(): boolean;
    /**
     * Cached snapshot state, to avoid unnecessary deep clones on repeated snapshot calls.
     * Cleared immediately (set to undefined) when the cache becomes invalid.
     */
    private snapshotCache;
    constructor(snapshot: QuorumClientsSnapshot);
    /**
     * Snapshots the current state of the QuorumClients
     * @returns a snapshot of the clients in the quorum
     */
    snapshot(): QuorumClientsSnapshot;
    /**
     * Adds a new client to the quorum
     */
    addMember(clientId: string, details: ISequencedClient): void;
    /**
     * Removes a client from the quorum
     */
    removeMember(clientId: string): void;
    /**
     * Retrieves all the members in the quorum
     */
    getMembers(): Map<string, ISequencedClient>;
    /**
     * Retrieves a specific member of the quorum
     */
    getMember(clientId: string): ISequencedClient | undefined;
    dispose(): void;
}
/**
 * The QuorumProposals holds a key/value store.  Proposed values become finalized in the store once all connected
 * clients have seen the proposal.
 * @internal
 */
export declare class QuorumProposals extends TypedEventEmitter<IQuorumProposals["on"]> implements IQuorumProposals {
    private readonly sendProposal;
    private readonly proposals;
    private readonly values;
    private isDisposed;
    get disposed(): boolean;
    private readonly stateEvents;
    /**
     * Cached snapshot state, to avoid unnecessary deep clones on repeated snapshot calls.
     * Cleared immediately (set to undefined) when the cache becomes invalid.
     */
    private proposalsSnapshotCache;
    private valuesSnapshotCache;
    constructor(snapshot: QuorumProposalsSnapshot, sendProposal: (key: string, value: unknown) => number);
    /**
     * Snapshots the current state of the QuorumProposals
     * @returns arrays of proposals and values
     */
    snapshot(): QuorumProposalsSnapshot;
    /**
     * Returns whether the quorum has achieved a consensus for the given key.
     */
    has(key: string): boolean;
    /**
     * Returns the consensus value for the given key
     */
    get(key: string): unknown;
    /**
     * Proposes a new value. Returns a promise that will either:
     * - Resolve when the proposal is accepted
     * - Reject if the proposal fails to send or if the QuorumProposals is disposed
     */
    propose(key: string, value: unknown): Promise<void>;
    /**
     * Begins tracking a new proposal
     */
    addProposal(key: string, value: unknown, sequenceNumber: number, local: boolean, clientSequenceNumber: number): void;
    /**
     * Updates the minimum sequence number. If the MSN advances past the sequence number for any proposal then it
     * becomes an approved value.
     */
    updateMinimumSequenceNumber(message: ISequencedDocumentMessage): void;
    setConnectionState(connected: boolean): void;
    dispose(): void;
}
/**
 * A quorum represents all clients currently within the collaboration window. As well as the values
 * they have agreed upon and any pending proposals.
 * @internal
 */
export declare class Quorum extends TypedEventEmitter<IQuorum["on"]> implements IQuorum {
    private readonly quorumClients;
    private readonly quorumProposals;
    private readonly isDisposed;
    get disposed(): boolean;
    constructor(members: QuorumClientsSnapshot, proposals: QuorumProposalsSnapshot["proposals"], values: QuorumProposalsSnapshot["values"], sendProposal: (key: string, value: unknown) => number);
    close(): void;
    /**
     * Snapshots the entire quorum
     * @returns a quorum snapshot
     */
    snapshot(): IQuorumSnapshot;
    /**
     * Returns whether the quorum has achieved a consensus for the given key.
     */
    has(key: string): boolean;
    /**
     * Returns the consensus value for the given key
     */
    get(key: string): unknown;
    /**
     * Adds a new client to the quorum
     */
    addMember(clientId: string, details: ISequencedClient): void;
    /**
     * Removes a client from the quorum
     */
    removeMember(clientId: string): void;
    /**
     * Retrieves all the members in the quorum
     */
    getMembers(): Map<string, ISequencedClient>;
    /**
     * Retrieves a specific member of the quorum
     */
    getMember(clientId: string): ISequencedClient | undefined;
    /**
     * Proposes a new value. Returns a promise that will resolve when the proposal is either accepted, or reject if
     * the proposal fails to send.
     */
    propose(key: string, value: unknown): Promise<void>;
    /**
     * Begins tracking a new proposal
     */
    addProposal(key: string, value: unknown, sequenceNumber: number, local: boolean, clientSequenceNumber: number): void;
    /**
     * Updates the minimum sequence number. If the MSN advances past the sequence number for any proposal then it
     * becomes an approved value.
     */
    updateMinimumSequenceNumber(message: ISequencedDocumentMessage): void;
    setConnectionState(connected: boolean, clientId?: string): void;
    dispose(): void;
}
//# sourceMappingURL=quorum.d.ts.map