/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { ISummaryTree } from "@fluidframework/driver-definitions";
import { IDocumentAttributes, ISnapshotTree } from "@fluidframework/driver-definitions/internal";
import { IDocumentStorageService, type ISnapshot } from "@fluidframework/driver-definitions/internal";
import { CombinedAppAndProtocolSummary, DeltaStreamConnectionForbiddenError } from "@fluidframework/driver-utils/internal";
import { ISerializableBlobContents } from "./containerStorageAdapter.js";
import type { IPendingContainerState, IPendingDetachedContainerState, ISnapshotInfo, SnapshotWithBlobs } from "./serializedStateManager.js";
export interface ISnapshotTreeWithBlobContents extends ISnapshotTree {
    blobsContents: {
        [path: string]: ArrayBufferLike;
    };
    trees: {
        [path: string]: ISnapshotTreeWithBlobContents;
    };
}
/**
 * Interface to represent the parsed parts of IResolvedUrl.url to help
 * in getting info about different parts of the url.
 * May not be compatible or relevant for any Url Resolver
 * @legacy
 * @alpha
 */
export interface IParsedUrl {
    /**
     * It is combination of tenantid/docId part of the url.
     */
    id: string;
    /**
     * It is the deep link path in the url.
     */
    path: string;
    /**
     * Query string part of the url.
     */
    query: string;
    /**
     * Undefined means load latest snapshot, otherwise it's version ID passed to IDocumentStorageService.getVersions()
     * to figure out what snapshot to use.
     */
    version: string | undefined;
}
/**
 * Utility api to parse the IResolvedUrl.url into specific parts like querystring, path to get
 * deep link info etc.
 * Warning - This function may not be compatible with any Url Resolver's resolved url. It works
 * with urls of type: protocol://<string>/.../..?<querystring>
 * @param url - This is the IResolvedUrl.url part of the resolved url.
 * @returns The IParsedUrl representing the input URL, or undefined if the format was not supported
 * @legacy
 * @alpha
 */
export declare function tryParseCompatibleResolvedUrl(url: string): IParsedUrl | undefined;
/**
 * Combine the app summary and protocol summary in 1 tree.
 * @param appSummary - Summary of the app.
 * @param protocolSummary - Summary of the protocol.
 * @internal
 */
export declare function combineAppAndProtocolSummary(appSummary: ISummaryTree, protocolSummary: ISummaryTree): CombinedAppAndProtocolSummary;
/**
 * Converts a snapshot to snapshotInfo with its blob contents
 * to align detached container format with IPendingContainerState
 *
 * Note, this assumes the ISnapshot sequence number is defined. Otherwise an assert will be thrown
 * @param snapshot - ISnapshot
 */
export declare function convertSnapshotToSnapshotInfo(snapshot: ISnapshot): ISnapshotInfo;
/**
 * Converts a snapshot to snapshotInfo with its blob contents
 * to align detached container format with IPendingContainerState
 *
 * Note, this assumes the ISnapshot sequence number is defined. Otherwise an assert will be thrown
 * @param snapshot - ISnapshot
 */
export declare function convertSnapshotInfoToSnapshot(snapshotInfo: ISnapshotInfo, snapshotSequenceNumber: number): ISnapshot;
export declare const getSnapshotTreeAndBlobsFromSerializedContainer: (detachedContainerSnapshot: ISummaryTree) => SnapshotWithBlobs;
export declare function getProtocolSnapshotTree(snapshot: ISnapshotTree): ISnapshotTree;
export declare const combineSnapshotTreeAndSnapshotBlobs: (baseSnapshot: ISnapshotTree, snapshotBlobs: ISerializableBlobContents) => ISnapshotTreeWithBlobContents;
export declare function isDeltaStreamConnectionForbiddenError(error: unknown): error is DeltaStreamConnectionForbiddenError;
/**
 * Parses the given string into {@link IPendingDetachedContainerState} format,
 * with validation (if invalid, throws a UsageError).
 * This is the inverse of the JSON.stringify call in {@link Container.serialize}
 */
export declare function getDetachedContainerStateFromSerializedContainer(serializedContainer: string): IPendingDetachedContainerState;
/**
 * Blindly parses the given string into {@link IPendingContainerState} format.
 * This is the inverse of the JSON.stringify call in {@link SerializedStateManager.getPendingLocalState}
 */
export declare function getAttachedContainerStateFromSerializedContainer(serializedContainer: string | undefined): IPendingContainerState | undefined;
/**
 * Ensures only a single instance of the provided async function is running.
 * If there are multiple calls they will all get the same promise to wait on.
 */
export declare const runSingle: <A extends any[], R>(func: (...args: A) => Promise<R>) => ((...args: A) => Promise<R>);
export declare function getDocumentAttributes(storage: Pick<IDocumentStorageService, "readBlob">, tree: ISnapshotTree | undefined): Promise<IDocumentAttributes>;
//# sourceMappingURL=utils.d.ts.map