"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDocumentAttributes = exports.runSingle = exports.getAttachedContainerStateFromSerializedContainer = exports.getDetachedContainerStateFromSerializedContainer = exports.isDeltaStreamConnectionForbiddenError = exports.combineSnapshotTreeAndSnapshotBlobs = exports.getProtocolSnapshotTree = exports.getSnapshotTreeAndBlobsFromSerializedContainer = exports.convertSnapshotInfoToSnapshot = exports.convertSnapshotToSnapshotInfo = exports.combineAppAndProtocolSummary = exports.tryParseCompatibleResolvedUrl = void 0;
const client_utils_1 = require("@fluid-internal/client-utils");
const internal_1 = require("@fluidframework/core-utils/internal");
const driver_definitions_1 = require("@fluidframework/driver-definitions");
const internal_2 = require("@fluidframework/driver-definitions/internal");
const internal_3 = require("@fluidframework/driver-utils/internal");
const internal_4 = require("@fluidframework/telemetry-utils/internal");
const uuid_1 = require("uuid");
/**
 * Utility api to parse the IResolvedUrl.url into specific parts like querystring, path to get
 * deep link info etc.
 * Warning - This function may not be compatible with any Url Resolver's resolved url. It works
 * with urls of type: protocol://<string>/.../..?<querystring>
 * @param url - This is the IResolvedUrl.url part of the resolved url.
 * @returns The IParsedUrl representing the input URL, or undefined if the format was not supported
 * @legacy
 * @alpha
 */
function tryParseCompatibleResolvedUrl(url) {
    const parsed = new URL(url);
    if (typeof parsed.pathname !== "string") {
        throw new internal_4.LoggingError("Failed to parse pathname");
    }
    const query = parsed.search ?? "";
    const regex = /^\/([^/]*\/[^/]*)(\/?.*)$/;
    const match = regex.exec(parsed.pathname);
    return match?.length === 3
        ? {
            id: match[1],
            path: match[2],
            query,
            // URLSearchParams returns null if the param is not provided.
            version: parsed.searchParams.get("version") ?? undefined,
        }
        : undefined;
}
exports.tryParseCompatibleResolvedUrl = tryParseCompatibleResolvedUrl;
/**
 * Combine the app summary and protocol summary in 1 tree.
 * @param appSummary - Summary of the app.
 * @param protocolSummary - Summary of the protocol.
 * @internal
 */
function combineAppAndProtocolSummary(appSummary, protocolSummary) {
    (0, internal_1.assert)(!(0, internal_3.isCombinedAppAndProtocolSummary)(appSummary), 0x5a8 /* app summary is already a combined tree! */);
    (0, internal_1.assert)(!(0, internal_3.isCombinedAppAndProtocolSummary)(protocolSummary), 0x5a9 /* protocol summary is already a combined tree! */);
    const createNewSummary = {
        type: driver_definitions_1.SummaryType.Tree,
        tree: {
            ".protocol": protocolSummary,
            ".app": appSummary,
        },
    };
    return createNewSummary;
}
exports.combineAppAndProtocolSummary = combineAppAndProtocolSummary;
/**
 * Converts a summary to snapshot tree and separate its blob contents
 * to align detached container format with IPendingContainerState
 * @param summary - ISummaryTree
 */
function convertSummaryToSnapshotAndBlobs(summary) {
    let blobContents = {};
    const treeNode = {
        blobs: {},
        trees: {},
        id: (0, uuid_1.v4)(),
        unreferenced: summary.unreferenced,
        groupId: summary.groupId,
    };
    for (const [key, summaryObject] of Object.entries(summary.tree)) {
        switch (summaryObject.type) {
            case driver_definitions_1.SummaryType.Tree: {
                const innerSnapshot = convertSummaryToSnapshotAndBlobs(summaryObject);
                treeNode.trees[key] = innerSnapshot.baseSnapshot;
                blobContents = { ...blobContents, ...innerSnapshot.snapshotBlobs };
                break;
            }
            case driver_definitions_1.SummaryType.Attachment: {
                treeNode.blobs[key] = summaryObject.id;
                break;
            }
            case driver_definitions_1.SummaryType.Blob: {
                const blobId = (0, uuid_1.v4)();
                treeNode.blobs[key] = blobId;
                const contentString = summaryObject.content instanceof Uint8Array
                    ? (0, client_utils_1.Uint8ArrayToString)(summaryObject.content)
                    : summaryObject.content;
                blobContents[blobId] = contentString;
                break;
            }
            case driver_definitions_1.SummaryType.Handle: {
                throw new internal_4.LoggingError("No handles should be there in summary in detached container!!");
            }
            default: {
                // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-explicit-any
                (0, internal_1.unreachableCase)(summaryObject, `Unknown tree type ${summaryObject.type}`);
            }
        }
    }
    const pendingSnapshot = { baseSnapshot: treeNode, snapshotBlobs: blobContents };
    return pendingSnapshot;
}
/**
 * Converts a snapshot to snapshotInfo with its blob contents
 * to align detached container format with IPendingContainerState
 *
 * Note, this assumes the ISnapshot sequence number is defined. Otherwise an assert will be thrown
 * @param snapshot - ISnapshot
 */
function convertSnapshotToSnapshotInfo(snapshot) {
    (0, internal_1.assert)(snapshot.sequenceNumber !== undefined, 0x93a /* Snapshot sequence number is missing */);
    const snapshotBlobs = {};
    for (const [blobId, arrayBufferLike] of snapshot.blobContents.entries()) {
        snapshotBlobs[blobId] = (0, client_utils_1.bufferToString)(arrayBufferLike, "utf8");
    }
    return {
        baseSnapshot: snapshot.snapshotTree,
        snapshotBlobs,
        snapshotSequenceNumber: snapshot.sequenceNumber,
    };
}
exports.convertSnapshotToSnapshotInfo = convertSnapshotToSnapshotInfo;
/**
 * Converts a snapshot to snapshotInfo with its blob contents
 * to align detached container format with IPendingContainerState
 *
 * Note, this assumes the ISnapshot sequence number is defined. Otherwise an assert will be thrown
 * @param snapshot - ISnapshot
 */
function convertSnapshotInfoToSnapshot(snapshotInfo, snapshotSequenceNumber) {
    const blobContents = new Map();
    for (const [blobId, serializedContent] of Object.entries(snapshotInfo.snapshotBlobs)) {
        blobContents.set(blobId, (0, client_utils_1.stringToBuffer)(serializedContent, "utf8"));
    }
    return {
        snapshotTree: snapshotInfo.baseSnapshot,
        blobContents,
        ops: [],
        sequenceNumber: snapshotSequenceNumber,
        latestSequenceNumber: undefined,
        snapshotFormatV: 1,
    };
}
exports.convertSnapshotInfoToSnapshot = convertSnapshotInfoToSnapshot;
/**
 * Converts summary parts into a SnapshotTree and its blob contents.
 * @param protocolSummaryTree - Protocol Summary Tree
 * @param appSummaryTree - App Summary Tree
 */
function convertProtocolAndAppSummaryToSnapshotAndBlobs(protocolSummaryTree, appSummaryTree) {
    const combinedSummary = {
        type: driver_definitions_1.SummaryType.Tree,
        tree: { ...appSummaryTree.tree },
    };
    combinedSummary.tree[".protocol"] = protocolSummaryTree;
    const snapshotTreeWithBlobContents = convertSummaryToSnapshotAndBlobs(combinedSummary);
    return snapshotTreeWithBlobContents;
}
const getSnapshotTreeAndBlobsFromSerializedContainer = (detachedContainerSnapshot) => {
    (0, internal_1.assert)((0, internal_3.isCombinedAppAndProtocolSummary)(detachedContainerSnapshot), 0x8e6 /* Protocol and App summary trees should be present */);
    const protocolSummaryTree = detachedContainerSnapshot.tree[".protocol"];
    const appSummaryTree = detachedContainerSnapshot.tree[".app"];
    const snapshotTreeWithBlobContents = convertProtocolAndAppSummaryToSnapshotAndBlobs(protocolSummaryTree, appSummaryTree);
    return snapshotTreeWithBlobContents;
};
exports.getSnapshotTreeAndBlobsFromSerializedContainer = getSnapshotTreeAndBlobsFromSerializedContainer;
function getProtocolSnapshotTree(snapshot) {
    return ".protocol" in snapshot.trees ? snapshot.trees[".protocol"] : snapshot;
}
exports.getProtocolSnapshotTree = getProtocolSnapshotTree;
const combineSnapshotTreeAndSnapshotBlobs = (baseSnapshot, snapshotBlobs) => {
    const blobsContents = {};
    // Process blobs in the current level
    for (const [, id] of Object.entries(baseSnapshot.blobs)) {
        if (snapshotBlobs[id] !== undefined) {
            blobsContents[id] = (0, client_utils_1.stringToBuffer)(snapshotBlobs[id], "utf8");
        }
    }
    // Recursively process trees in the current level
    const trees = {};
    for (const [path, tree] of Object.entries(baseSnapshot.trees)) {
        trees[path] = (0, exports.combineSnapshotTreeAndSnapshotBlobs)(tree, snapshotBlobs);
    }
    // Create a new snapshot tree with blob contents and processed trees
    const snapshotTreeWithBlobContents = {
        ...baseSnapshot,
        blobsContents,
        trees,
    };
    return snapshotTreeWithBlobContents;
};
exports.combineSnapshotTreeAndSnapshotBlobs = combineSnapshotTreeAndSnapshotBlobs;
function isDeltaStreamConnectionForbiddenError(error) {
    return (typeof error === "object" &&
        error !== null &&
        error?.errorType ===
            internal_2.DriverErrorTypes.deltaStreamConnectionForbidden);
}
exports.isDeltaStreamConnectionForbiddenError = isDeltaStreamConnectionForbiddenError;
/**
 * Validates format in parsed string get from detached container
 * serialization using IPendingDetachedContainerState format.
 */
function isPendingDetachedContainerState(detachedContainerState) {
    if (detachedContainerState?.attached === undefined ||
        detachedContainerState?.baseSnapshot === undefined ||
        detachedContainerState?.snapshotBlobs === undefined ||
        detachedContainerState?.hasAttachmentBlobs === undefined) {
        return false;
    }
    return true;
}
/**
 * Parses the given string into {@link IPendingDetachedContainerState} format,
 * with validation (if invalid, throws a UsageError).
 * This is the inverse of the JSON.stringify call in {@link Container.serialize}
 */
function getDetachedContainerStateFromSerializedContainer(serializedContainer) {
    const hasBlobsSummaryTree = ".hasAttachmentBlobs";
    // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
    const parsedContainerState = JSON.parse(serializedContainer);
    // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
    if (isPendingDetachedContainerState(parsedContainerState)) {
        return parsedContainerState;
        // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
    }
    else if ((0, internal_3.isCombinedAppAndProtocolSummary)(parsedContainerState)) {
        const { baseSnapshot, snapshotBlobs } = (0, exports.getSnapshotTreeAndBlobsFromSerializedContainer)(parsedContainerState);
        const detachedContainerState = {
            attached: false,
            baseSnapshot,
            snapshotBlobs,
            hasAttachmentBlobs: parsedContainerState.tree[hasBlobsSummaryTree] !== undefined,
        };
        return detachedContainerState;
    }
    else {
        throw new internal_4.UsageError("Cannot rehydrate detached container. Incorrect format");
    }
}
exports.getDetachedContainerStateFromSerializedContainer = getDetachedContainerStateFromSerializedContainer;
/**
 * Blindly parses the given string into {@link IPendingContainerState} format.
 * This is the inverse of the JSON.stringify call in {@link SerializedStateManager.getPendingLocalState}
 */
function getAttachedContainerStateFromSerializedContainer(serializedContainer) {
    return serializedContainer === undefined
        ? undefined
        : JSON.parse(serializedContainer);
}
exports.getAttachedContainerStateFromSerializedContainer = getAttachedContainerStateFromSerializedContainer;
/**
 * Ensures only a single instance of the provided async function is running.
 * If there are multiple calls they will all get the same promise to wait on.
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
const runSingle = (func) => {
    let running;
    // don't mark this function async, so we return the same promise,
    // rather than one that is wrapped due to async
    // eslint-disable-next-line @typescript-eslint/promise-function-async
    return (...args) => {
        if (running !== undefined) {
            if (!(0, internal_1.compareArrays)(running.args, args)) {
                return Promise.reject(new internal_4.UsageError("Subsequent calls cannot use different arguments."));
            }
            return running.result;
        }
        running = { args, result: func(...args).finally(() => (running = undefined)) };
        return running.result;
    };
};
exports.runSingle = runSingle;
async function getDocumentAttributes(storage, tree) {
    if (tree === undefined) {
        return {
            minimumSequenceNumber: 0,
            sequenceNumber: 0,
        };
    }
    // Backward compatibility: old docs would have ".attributes" instead of "attributes"
    const attributesHash = ".protocol" in tree.trees
        ? tree.trees[".protocol"].blobs.attributes
        : tree.blobs[".attributes"];
    const attributes = await (0, internal_3.readAndParse)(storage, attributesHash);
    return attributes;
}
exports.getDocumentAttributes = getDocumentAttributes;
//# sourceMappingURL=utils.js.map