/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { ICriticalContainerError } from "@fluidframework/container-definitions";
import { IDeltaQueue, ReadOnlyInfo } from "@fluidframework/container-definitions/internal";
import { ITelemetryBaseProperties } from "@fluidframework/core-interfaces";
import { ConnectionMode, IClient, IClientDetails } from "@fluidframework/driver-definitions";
import { IDocumentService, IClientConfiguration, IDocumentMessage, ISequencedDocumentMessage } from "@fluidframework/driver-definitions/internal";
import { ITelemetryLoggerExt } from "@fluidframework/telemetry-utils/internal";
import { IConnectionDetailsInternal, IConnectionManager, IConnectionManagerFactoryArgs, IConnectionStateChangeReason, ReconnectMode } from "./contracts.js";
/**
 * Implementation of IConnectionManager, used by Container class
 * Implements constant connectivity to relay service, by reconnecting in case of lost connection or error.
 * Exposes various controls to influence this process, including manual reconnects, forced read-only mode, etc.
 */
export declare class ConnectionManager implements IConnectionManager {
    private readonly serviceProvider;
    readonly containerDirty: () => boolean;
    private readonly client;
    private readonly logger;
    private readonly props;
    /**
     * Connection mode used when reconnecting on error or disconnect.
     */
    private readonly defaultReconnectionMode;
    /**
     * Tracks the current in-progress connection attempt. Undefined if there is none.
     * Note: Once the connection attempt fires and the code becomes asynchronous, its possible that a new connection
     * attempt was fired and this.pendingConnection was overwritten to reflect the new attempt.
     */
    private pendingConnection;
    private connection;
    /**
     * Details about connection. undefined if there is no active connection.
     */
    private _connectionDetails?;
    /**
     * file ACL - whether user has only read-only access to a file
     */
    private _readonlyPermissions;
    /**
     * tracks host requiring read-only mode.
     */
    private _forceReadonly;
    /**
     * Controls whether the DeltaManager will automatically reconnect to the delta stream after receiving a disconnect.
     */
    private _reconnectMode;
    /**
     * True if there is pending (async) reconnection from "read" to "write"
     */
    private pendingReconnect;
    private clientSequenceNumber;
    private clientSequenceNumberObserved;
    /**
     * Counts the number of non-runtime ops sent by the client which may not be acked.
     */
    private localOpsToIgnore;
    /**
     * track clientId used last time when we sent any ops
     */
    private lastSubmittedClientId;
    private connectFirstConnection;
    private _connectionVerboseProps;
    private _connectionProps;
    private _disposed;
    private readonly _outbound;
    get connectionVerboseProps(): Record<string, string | number>;
    readonly clientDetails: IClientDetails;
    /**
     * The current connection mode, initially read.
     */
    get connectionMode(): ConnectionMode;
    get connected(): boolean;
    get clientId(): string | undefined;
    /**
     * Details about connection. Returns undefined if there is no active connection.
     */
    get connectionDetails(): IConnectionDetailsInternal | undefined;
    /**
     * Automatic reconnecting enabled or disabled.
     * If set to Never, then reconnecting will never be allowed.
     */
    get reconnectMode(): ReconnectMode;
    get maxMessageSize(): number;
    get version(): string;
    get serviceConfiguration(): IClientConfiguration | undefined;
    get scopes(): string[] | undefined;
    get outbound(): IDeltaQueue<IDocumentMessage[]>;
    /**
     * Returns set of props that can be logged in telemetry that provide some insights / statistics
     * about current or last connection (if there is no connection at the moment)
     */
    get connectionProps(): ITelemetryBaseProperties;
    shouldJoinWrite(): boolean;
    /**
     * Tells if container is in read-only mode.
     * Data stores should listen for "readonly" notifications and disallow user
     * making changes to data stores.
     * Readonly state can be because of no storage write permission,
     * or due to host forcing readonly mode for container.
     * It is undefined if we have not yet established websocket connection
     * and do not know if user has write access to a file.
     */
    private get readonly();
    get readOnlyInfo(): ReadOnlyInfo;
    private static detailsFromConnection;
    constructor(serviceProvider: () => IDocumentService | undefined, containerDirty: () => boolean, client: IClient, reconnectAllowed: boolean, logger: ITelemetryLoggerExt, props: IConnectionManagerFactoryArgs);
    dispose(error?: ICriticalContainerError, switchToReadonly?: boolean): void;
    /**
     * Enables or disables automatic reconnecting.
     * Will throw an error if reconnectMode set to Never.
     */
    setAutoReconnect(mode: ReconnectMode, reason: IConnectionStateChangeReason): void;
    /**
     * {@inheritDoc Container.forceReadonly}
     */
    forceReadonly(readonly: boolean): void;
    private set_readonlyPermissions;
    connect(reason: IConnectionStateChangeReason, connectionMode?: ConnectionMode): void;
    private connectCore;
    /**
     * Start the connection. Any error should result in container being closed.
     * And report the error if it escapes for any reason.
     * @param args - The connection arguments
     */
    private triggerConnect;
    /**
     * Disconnect the current connection.
     * @param reason - Text description of disconnect reason to emit with disconnect event
     * @param error - Error causing the disconnect if any.
     * @returns A boolean that indicates if there was an existing connection (or pending connection) to disconnect
     */
    private disconnectFromDeltaStream;
    /**
     * Cancel in-progress connection attempt.
     */
    private cancelConnection;
    /**
     * Once we've successfully gotten a connection, we need to set up state, attach event listeners, and process
     * initial messages.
     * @param connection - The newly established connection
     */
    private setupNewSuccessfulConnection;
    /**
     * Disconnect the current connection and reconnect. Closes the container if it fails.
     * @param connection - The connection that wants to reconnect - no-op if it's different from this.connection
     * @param requestedMode - Read or write
     * @param error - Error reconnect information including whether or not to reconnect
     * @returns A promise that resolves when the connection is reestablished or we stop trying
     */
    private reconnectOnError;
    /**
     * Disconnect the current connection and reconnect.
     * @param connection - The connection that wants to reconnect - no-op if it's different from this.connection
     * @param requestedMode - Read or write
     * @param error - Error reconnect information including whether or not to reconnect
     * @returns A promise that resolves when the connection is reestablished or we stop trying
     */
    private reconnect;
    prepareMessageToSend(message: Omit<IDocumentMessage, "clientSequenceNumber">): IDocumentMessage | undefined;
    submitSignal(content: string, targetClientId?: string): void;
    sendMessages(messages: IDocumentMessage[]): void;
    beforeProcessingIncomingOp(message: ISequencedDocumentMessage): void;
    private readonly opHandler;
    private readonly signalHandler;
    private readonly nackHandler;
    private readonly disconnectHandlerInternal;
    private readonly errorHandler;
}
//# sourceMappingURL=connectionManager.d.ts.map