/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { ISnapshotTreeWithBlobContents } from "@fluidframework/container-definitions/internal";
import { IDisposable } from "@fluidframework/core-interfaces";
import { ISummaryHandle, ISummaryTree } from "@fluidframework/driver-definitions";
import { FetchSource, IDocumentService, IDocumentStorageService, IDocumentStorageServicePolicies, ISnapshot, ISnapshotFetchOptions, ISummaryContext, ICreateBlobResponse, ISnapshotTree, IVersion } from "@fluidframework/driver-definitions/internal";
import { ITelemetryLoggerExt } from "@fluidframework/telemetry-utils/internal";
import type { MemoryDetachedBlobStorage } from "./memoryBlobStorage.js";
import type { ISerializedStateManagerDocumentStorageService, ISnapshotInfo } from "./serializedStateManager.js";
/**
 * Stringified blobs from a summary/snapshot tree.
 * @internal
 */
export interface ISerializableBlobContents {
    [id: string]: string;
}
/**
 * This class wraps the actual storage and make sure no wrong apis are called according to
 * container attach state.
 */
export declare class ContainerStorageAdapter implements ISerializedStateManagerDocumentStorageService, IDocumentStorageService, IDisposable {
    private readonly logger;
    /**
     * ArrayBufferLikes or utf8 encoded strings, containing blobs from a snapshot
     */
    private readonly blobContents;
    private loadingGroupIdSnapshotsFromPendingState;
    private readonly addProtocolSummaryIfMissing;
    private readonly enableSummarizeProtocolTree;
    private _storageService;
    private _summarizeProtocolTree;
    /**
     * Whether the adapter will enforce sending combined summary trees.
     */
    get summarizeProtocolTree(): boolean;
    private _loadedGroupIdSnapshots;
    /**
     * Any loading group id (virtualized) snapshot download from storage will be stored here.
     */
    get loadedGroupIdSnapshots(): Record<string, ISnapshot>;
    /**
     * An adapter that ensures we're using detachedBlobStorage up until we connect to a real service, and then
     * after connecting to a real service augments it with retry and combined summary tree enforcement.
     * @param detachedBlobStorage - The detached blob storage to use up until we connect to a real service
     * @param logger - Telemetry logger
     * @param loadingGroupIdSnapshotsFromPendingState - in offline mode, any loading group snapshots we've downloaded from the service that were stored in the pending state
     * @param addProtocolSummaryIfMissing - a callback to permit the container to inspect the summary we're about to
     * upload, and fix it up with a protocol tree if needed
     * @param enableSummarizeProtocolTree - Enable uploading a protocol summary. Note: preference is given to service policy's "summarizeProtocolTree" before this value.
     */
    constructor(detachedBlobStorage: MemoryDetachedBlobStorage | undefined, logger: ITelemetryLoggerExt, 
    /**
     * ArrayBufferLikes or utf8 encoded strings, containing blobs from a snapshot
     */
    blobContents: {
        [id: string]: ArrayBufferLike | string;
    }, loadingGroupIdSnapshotsFromPendingState: Record<string, ISnapshotInfo> | undefined, addProtocolSummaryIfMissing: (summaryTree: ISummaryTree) => ISummaryTree, enableSummarizeProtocolTree: boolean | undefined);
    disposed: boolean;
    dispose(error?: Error): void;
    connectToService(service: IDocumentService): void;
    loadSnapshotFromSnapshotBlobs(snapshotBlobs: ISerializableBlobContents): void;
    clearPendingState(): void;
    get policies(): IDocumentStorageServicePolicies | undefined;
    getSnapshotTree(version?: IVersion, scenarioName?: string): Promise<ISnapshotTree | null>;
    getSnapshot(snapshotFetchOptions?: ISnapshotFetchOptions): Promise<ISnapshot>;
    readBlob(id: string): Promise<ArrayBufferLike>;
    getVersions(versionId: string | null, count: number, scenarioName?: string, fetchSource?: FetchSource): Promise<IVersion[]>;
    uploadSummaryWithContext(summary: ISummaryTree, context: ISummaryContext): Promise<string>;
    downloadSummary(handle: ISummaryHandle): Promise<ISummaryTree>;
    createBlob(file: ArrayBufferLike): Promise<ICreateBlobResponse>;
}
/**
 * Get blob contents of a snapshot tree from storage (or, ideally, cache)
 */
export declare function getBlobContentsFromTree(snapshot: ISnapshot | ISnapshotTree, storage: Pick<IDocumentStorageService, "readBlob">): Promise<ISerializableBlobContents>;
/**
 * Extract blob contents from a snapshot tree with blob contents
 */
export declare function getBlobContentsFromTreeWithBlobContents(snapshot: ISnapshotTreeWithBlobContents): ISerializableBlobContents;
//# sourceMappingURL=containerStorageAdapter.d.ts.map