/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { bufferToString, stringToBuffer } from "@fluid-internal/client-utils";
import { assert } from "@fluidframework/core-utils/internal";
import { isInstanceOfISnapshot, UsageError } from "@fluidframework/driver-utils/internal";
import { ProtocolTreeStorageService } from "./protocolTreeDocumentStorageService.js";
import { RetriableDocumentStorageService } from "./retriableDocumentStorageService.js";
import { convertSnapshotInfoToSnapshot } from "./utils.js";
/**
 * This class wraps the actual storage and make sure no wrong apis are called according to
 * container attach state.
 */
export class ContainerStorageAdapter {
    /**
     * Whether the adapter will enforce sending combined summary trees.
     */
    get summarizeProtocolTree() {
        return this._summarizeProtocolTree === true;
    }
    /**
     * Any loading group id (virtualized) snapshot download from storage will be stored here.
     */
    get loadedGroupIdSnapshots() {
        return this._loadedGroupIdSnapshots;
    }
    /**
     * An adapter that ensures we're using detachedBlobStorage up until we connect to a real service, and then
     * after connecting to a real service augments it with retry and combined summary tree enforcement.
     * @param detachedBlobStorage - The detached blob storage to use up until we connect to a real service
     * @param logger - Telemetry logger
     * @param loadingGroupIdSnapshotsFromPendingState - in offline mode, any loading group snapshots we've downloaded from the service that were stored in the pending state
     * @param addProtocolSummaryIfMissing - a callback to permit the container to inspect the summary we're about to
     * upload, and fix it up with a protocol tree if needed
     * @param enableSummarizeProtocolTree - Enable uploading a protocol summary. Note: preference is given to service policy's "summarizeProtocolTree" before this value.
     */
    constructor(detachedBlobStorage, logger, 
    /**
     * ArrayBufferLikes or utf8 encoded strings, containing blobs from a snapshot
     */
    blobContents = {}, loadingGroupIdSnapshotsFromPendingState, addProtocolSummaryIfMissing, enableSummarizeProtocolTree) {
        this.logger = logger;
        this.blobContents = blobContents;
        this.loadingGroupIdSnapshotsFromPendingState = loadingGroupIdSnapshotsFromPendingState;
        this.addProtocolSummaryIfMissing = addProtocolSummaryIfMissing;
        this.enableSummarizeProtocolTree = enableSummarizeProtocolTree;
        this._loadedGroupIdSnapshots = {};
        this.disposed = false;
        this._storageService = new BlobOnlyStorage(detachedBlobStorage, logger);
    }
    dispose(error) {
        this._storageService?.dispose?.(error);
        this.disposed = true;
    }
    connectToService(service) {
        if (!(this._storageService instanceof BlobOnlyStorage)) {
            return;
        }
        const storageServiceP = service.connectToStorage();
        const retriableStorage = (this._storageService = new RetriableDocumentStorageService(storageServiceP, this.logger));
        // A storage service wrapper which intercept calls to uploadSummaryWithContext and ensure they include
        // the protocol summary, provided single-commit summary is enabled.
        this._storageService = new ProtocolTreeStorageService(retriableStorage, (...props) => {
            this.logger.sendTelemetryEvent({ eventName: "summarizeProtocolTreeEnabled" });
            return this.addProtocolSummaryIfMissing(...props);
        }, 
        // A callback to ensure we fetch the most updated value of service.policies.summarizeProtocolTree, which could be set
        // based on the response received from the service after connection is established.
        () => {
            // Determine whether or not container should upload the protocol summary along with the summary.
            // This is determined based on what value is set for serve policy's summariProtocolTree value or the enableSummarizeProtocolTree
            // retrievd from the loader options or monitoring context config.
            const shouldSummarizeProtocolTree = service.policies?.summarizeProtocolTree ?? this.enableSummarizeProtocolTree ?? false;
            if (this._summarizeProtocolTree !== shouldSummarizeProtocolTree) {
                this.logger.sendTelemetryEvent({
                    eventName: "isSummarizeProtocolTreeEnabled",
                    details: { value: shouldSummarizeProtocolTree },
                });
            }
            this._summarizeProtocolTree = shouldSummarizeProtocolTree;
            return this._summarizeProtocolTree;
        });
    }
    loadSnapshotFromSnapshotBlobs(snapshotBlobs) {
        for (const [id, value] of Object.entries(snapshotBlobs)) {
            this.blobContents[id] = value;
        }
    }
    clearPendingState() {
        this.loadingGroupIdSnapshotsFromPendingState = undefined;
    }
    get policies() {
        // back-compat 0.40 containerRuntime requests policies even in detached container if storage is present
        // and storage is always present in >=0.41.
        try {
            return this._storageService.policies;
        }
        catch {
            // No-op
        }
        return undefined;
    }
    async getSnapshotTree(version, scenarioName) {
        return this._storageService.getSnapshotTree(version, scenarioName);
    }
    async getSnapshot(snapshotFetchOptions) {
        let snapshot;
        if (this.loadingGroupIdSnapshotsFromPendingState !== undefined &&
            snapshotFetchOptions?.loadingGroupIds !== undefined) {
            const localSnapshot = this.loadingGroupIdSnapshotsFromPendingState[snapshotFetchOptions.loadingGroupIds[0]];
            assert(localSnapshot !== undefined, 0x970 /* Local snapshot must be present */);
            snapshot = convertSnapshotInfoToSnapshot(localSnapshot, localSnapshot.snapshotSequenceNumber);
        }
        else {
            if (this._storageService.getSnapshot === undefined) {
                throw new UsageError("getSnapshot api should exist in internal storage in ContainerStorageAdapter");
            }
            snapshot = await this._storageService.getSnapshot(snapshotFetchOptions);
        }
        // Track the latest snapshot for each loading group id
        const loadingGroupIds = snapshotFetchOptions?.loadingGroupIds;
        assert(snapshot.sequenceNumber !== undefined, 0x971 /* Snapshot must have sequence number */);
        if (loadingGroupIds !== undefined) {
            for (const loadingGroupId of loadingGroupIds) {
                // Do we actually want to update the stored snapshot?
                // What if the incoming snapshot is way newer than the stored snapshot?
                // We only want to update the stored snapshot if the incoming snapshot is newer (stored sequence number < incoming sequence number)
                const storedSeqNum = this._loadedGroupIdSnapshots[loadingGroupId]?.sequenceNumber ?? -1;
                if (storedSeqNum < snapshot.sequenceNumber) {
                    this._loadedGroupIdSnapshots[loadingGroupId] = snapshot;
                }
            }
        }
        return snapshot;
    }
    async readBlob(id) {
        const maybeBlob = this.blobContents[id];
        if (maybeBlob !== undefined) {
            if (typeof maybeBlob === "string") {
                const blob = stringToBuffer(maybeBlob, "utf8");
                return blob;
            }
            return maybeBlob;
        }
        return this._storageService.readBlob(id);
    }
    async getVersions(
    // API used below uses null
    // eslint-disable-next-line @rushstack/no-new-null
    versionId, count, scenarioName, fetchSource) {
        return this._storageService.getVersions(versionId, count, scenarioName, fetchSource);
    }
    async uploadSummaryWithContext(summary, context) {
        return this._storageService.uploadSummaryWithContext(summary, context);
    }
    async downloadSummary(handle) {
        return this._storageService.downloadSummary(handle);
    }
    async createBlob(file) {
        return this._storageService.createBlob(file);
    }
}
/**
 * Storage which only supports createBlob() and readBlob(). This is used with IDetachedBlobStorage to support
 * blobs in detached containers.
 */
class BlobOnlyStorage {
    constructor(detachedStorage, logger) {
        this.detachedStorage = detachedStorage;
        this.logger = logger;
        /* eslint-disable @typescript-eslint/unbound-method */
        // eslint-disable-next-line @rushstack/no-new-null
        this.getSnapshotTree = this.notCalled;
        this.getSnapshot = this.notCalled;
        this.getVersions = this.notCalled;
        this.write = this.notCalled;
        this.uploadSummaryWithContext = this.notCalled;
        this.downloadSummary = this.notCalled;
    }
    async createBlob(content) {
        return this.verifyStorage().createBlob(content);
    }
    async readBlob(blobId) {
        return this.verifyStorage().readBlob(blobId);
    }
    verifyStorage() {
        if (this.detachedStorage === undefined) {
            throw new UsageError("Real storage calls not allowed in Unattached container");
        }
        return this.detachedStorage;
    }
    get policies() {
        return this.notCalled();
    }
    /* eslint-enable @typescript-eslint/unbound-method */
    notCalled() {
        this.verifyStorage();
        try {
            // some browsers may not populate stack unless exception is thrown
            throw new Error("BlobOnlyStorage not implemented method used");
        }
        catch (error) {
            this.logger.sendTelemetryEvent({ eventName: "BlobOnlyStorageWrongCall" }, error);
            throw error;
        }
    }
}
// runtime will write a tree to the summary containing "attachment" type entries
// which reference attachment blobs by ID, along with a blob containing the blob redirect table.
// However, some drivers do not support the "attachment" type and will convert them to "blob" type
// entries. We want to avoid saving these to reduce the size of stashed change blobs, but we
// need to make sure the blob redirect table is saved.
const blobsTreeName = ".blobs";
const redirectTableBlobName = ".redirectTable";
/**
 * Get blob contents of a snapshot tree from storage (or, ideally, cache)
 */
export async function getBlobContentsFromTree(snapshot, storage) {
    const blobs = {};
    if (isInstanceOfISnapshot(snapshot)) {
        const blobContents = snapshot.blobContents;
        for (const [id, content] of blobContents.entries()) {
            // ArrayBufferLike will not survive JSON.stringify()
            blobs[id] = bufferToString(content, "utf8");
        }
    }
    else {
        await getBlobContentsFromTreeCore(snapshot, blobs, storage);
    }
    return blobs;
}
async function getBlobContentsFromTreeCore(tree, blobs, storage, root = true) {
    const treePs = [];
    for (const [key, subTree] of Object.entries(tree.trees)) {
        if (root && key === blobsTreeName) {
            treePs.push(getBlobManagerTreeFromTree(subTree, blobs, storage));
        }
        else {
            treePs.push(getBlobContentsFromTreeCore(subTree, blobs, storage, false));
        }
    }
    for (const id of Object.values(tree.blobs)) {
        const blob = await storage.readBlob(id);
        // ArrayBufferLike will not survive JSON.stringify()
        blobs[id] = bufferToString(blob, "utf8");
    }
    return Promise.all(treePs);
}
// save redirect table from .blobs tree but nothing else
async function getBlobManagerTreeFromTree(tree, blobs, storage) {
    const id = tree.blobs[redirectTableBlobName];
    assert(id !== undefined, 0x9ce /* id is undefined in getBlobManagerTreeFromTree */);
    const blob = await storage.readBlob(id);
    // ArrayBufferLike will not survive JSON.stringify()
    blobs[id] = bufferToString(blob, "utf8");
}
/**
 * Extract blob contents from a snapshot tree with blob contents
 */
export function getBlobContentsFromTreeWithBlobContents(snapshot) {
    const blobs = {};
    getBlobContentsFromTreeWithBlobContentsCore(snapshot, blobs);
    return blobs;
}
function getBlobContentsFromTreeWithBlobContentsCore(tree, blobs, root = true) {
    for (const [key, subTree] of Object.entries(tree.trees)) {
        if (root && key === blobsTreeName) {
            getBlobManagerTreeFromTreeWithBlobContents(subTree, blobs);
        }
        else {
            getBlobContentsFromTreeWithBlobContentsCore(subTree, blobs, false);
        }
    }
    for (const id of Object.values(tree.blobs)) {
        const blob = tree.blobsContents?.[id];
        assert(blob !== undefined, 0x2ec /* "Blob must be present in blobsContents" */);
        // ArrayBufferLike will not survive JSON.stringify()
        blobs[id] = bufferToString(blob, "utf8");
    }
}
// save redirect table from .blobs tree but nothing else
function getBlobManagerTreeFromTreeWithBlobContents(tree, blobs) {
    const id = tree.blobs[redirectTableBlobName];
    assert(id !== undefined, 0x9cf /* id is undefined in getBlobManagerTreeFromTreeWithBlobContents */);
    const blob = tree.blobsContents?.[id];
    assert(blob !== undefined, 0x70f /* Blob must be present in blobsContents */);
    // ArrayBufferLike will not survive JSON.stringify()
    blobs[id] = bufferToString(blob, "utf8");
}
//# sourceMappingURL=containerStorageAdapter.js.map