/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { ICriticalContainerError } from "@fluidframework/container-definitions";
import { IDeltaManagerEvents, IDeltaManagerFull, IDeltaQueue, type IDeltaSender, type ReadOnlyInfo } from "@fluidframework/container-definitions/internal";
import { IEventProvider, ITelemetryBaseProperties } from "@fluidframework/core-interfaces";
import { IThrottlingWarning } from "@fluidframework/core-interfaces/internal";
import { ConnectionMode } from "@fluidframework/driver-definitions";
import { IDocumentService, IDocumentMessage, MessageType, ISequencedDocumentMessage, ISignalMessage, type IClientDetails, type IClientConfiguration } from "@fluidframework/driver-definitions/internal";
import { type ITelemetryErrorEventExt, ITelemetryLoggerExt, EventEmitterWithErrorHandling } from "@fluidframework/telemetry-utils/internal";
import { IConnectionDetailsInternal, IConnectionManager, IConnectionManagerFactoryArgs, IConnectionStateChangeReason } from "./contracts.js";
export interface IConnectionArgs {
    mode?: ConnectionMode;
    fetchOpsFromStorage?: boolean;
    reason: IConnectionStateChangeReason;
}
/**
 * Includes events emitted by the concrete implementation DeltaManager
 * but not exposed on the public interface IDeltaManager
 */
export interface IDeltaManagerInternalEvents extends IDeltaManagerEvents {
    (event: "throttled", listener: (error: IThrottlingWarning) => void): any;
    (event: "closed" | "disposed", listener: (error?: ICriticalContainerError) => void): any;
    (event: "connect", listener: (details: IConnectionDetailsInternal, opsBehind?: number) => void): any;
    (event: "establishingConnection", listener: (reason: IConnectionStateChangeReason) => void): any;
    (event: "cancelEstablishingConnection", listener: (reason: IConnectionStateChangeReason) => void): any;
}
/**
 * Interface used to define a strategy for handling incoming delta messages
 */
export interface IDeltaHandlerStrategy {
    /**
     * Processes the message.
     */
    process: (message: ISequencedDocumentMessage) => void;
    /**
     * Processes the signal.
     */
    processSignal: (message: ISignalMessage) => void;
}
/**
 * Manages the flow of both inbound and outbound messages. This class ensures that shared objects receive delta
 * messages in order regardless of possible network conditions or timings causing out of order delivery.
 */
export declare class DeltaManager<TConnectionManager extends IConnectionManager> extends EventEmitterWithErrorHandling<IDeltaManagerInternalEvents> implements IDeltaManagerFull, IEventProvider<IDeltaManagerInternalEvents> {
    private readonly serviceProvider;
    private readonly logger;
    private readonly _active;
    readonly connectionManager: TConnectionManager;
    get active(): boolean;
    get disposed(): boolean;
    get IDeltaSender(): IDeltaSender;
    private pending;
    private fetchReason;
    private currentlyProcessingOps;
    private minSequenceNumber;
    private lastQueuedSequenceNumber;
    private lastObservedSeqNumber;
    private lastProcessedSequenceNumber;
    private lastProcessedMessage;
    /**
     * Count the number of noops sent by the client which may not be acked
     */
    private noOpCount;
    /**
     * Track clientSequenceNumber of the last op
     */
    private lastClientSequenceNumber;
    /**
     * Track down the ops size.
     */
    private opsSize;
    private prevEnqueueMessagesReason;
    private previouslyProcessedMessage;
    private initSequenceNumber;
    private readonly _inbound;
    private readonly _inboundSignal;
    private _closed;
    private _disposed;
    private handler;
    private deltaStorage;
    private readonly throttlingIdSet;
    private timeTillThrottling;
    readonly closeAbortController: AbortController;
    private readonly deltaStorageDelayId;
    private readonly deltaStreamDelayId;
    private messageBuffer;
    private _checkpointSequenceNumber;
    get inbound(): IDeltaQueue<ISequencedDocumentMessage>;
    get inboundSignal(): IDeltaQueue<ISignalMessage>;
    get initialSequenceNumber(): number;
    get lastSequenceNumber(): number;
    get lastMessage(): ISequencedDocumentMessage | undefined;
    get lastKnownSeqNumber(): number;
    get minimumSequenceNumber(): number;
    /**
     * Tells if  current connection has checkpoint information.
     * I.e. we know how far behind the client was at the time of establishing connection
     */
    get hasCheckpointSequenceNumber(): boolean;
    get maxMessageSize(): number;
    get version(): string;
    get serviceConfiguration(): IClientConfiguration | undefined;
    get outbound(): IDeltaQueue<IDocumentMessage[]>;
    get readOnlyInfo(): ReadOnlyInfo;
    get clientDetails(): IClientDetails;
    submit(type: MessageType, contents?: string, batch?: boolean, metadata?: unknown, compression?: string, referenceSequenceNumber?: number): number;
    submitSignal(content: string, targetClientId?: string): void;
    flush(): void;
    get connectionProps(): ITelemetryBaseProperties;
    /**
     * Log error event with a bunch of internal to DeltaManager information about state of op processing
     * Used to diagnose connectivity issues related to op processing (i.e. cases where for some reason
     * we stop processing ops that results in no processing join op and thus moving to connected state)
     * @param event - Event to log.
     */
    logConnectionIssue(event: ITelemetryErrorEventExt): void;
    constructor(serviceProvider: () => IDocumentService | undefined, logger: ITelemetryLoggerExt, _active: () => boolean, createConnectionManager: (props: IConnectionManagerFactoryArgs) => TConnectionManager);
    private cancelEstablishingConnection;
    private establishingConnection;
    private connectHandler;
    /**
     * Sets the sequence number from which inbound messages should be returned
     * @param snapshotSequenceNumber - The sequence number of the snapshot at which the document loaded from.
     * @param lastProcessedSequenceNumber - The last processed sequence number, for offline, it should be greater than the sequence number.
     * Setting lastProcessedSequenceNumber allows the DeltaManager to skip downloading and processing ops that have already been processed.
     */
    attachOpHandler(minSequenceNumber: number, snapshotSequenceNumber: number, handler: IDeltaHandlerStrategy, prefetchType?: "cached" | "all" | "none", lastProcessedSequenceNumber?: number): Promise<void>;
    connect(args: IConnectionArgs): void;
    private getDeltas;
    /**
     * Closes the connection and clears inbound & outbound queues.
     *
     * Differences from dispose:
     * - close will trigger readonly notification
     * - close emits "closed"
     * - close cannot be called after dispose
     */
    close(error?: ICriticalContainerError): void;
    /**
     * Disposes the connection and clears the inbound & outbound queues.
     *
     * Differences from close:
     * - dispose will emit "disposed"
     * - dispose will remove all listeners
     * - dispose can be called after closure
     */
    dispose(error?: Error | ICriticalContainerError): void;
    private clearQueues;
    refreshDelayInfo(id: string): void;
    private disconnectHandler;
    /**
     * Emit info about a delay in service communication on account of throttling.
     * @param id - Id of the connection that is delayed
     * @param delayMs - Duration of the delay
     * @param error - error object indicating the throttling
     */
    emitDelayInfo(id: string, delayMs: number, error: unknown): void;
    private comparableMessagePayload;
    private enqueueMessages;
    private processInboundMessage;
    /**
     * Retrieves the missing deltas between the given sequence numbers
     */
    private fetchMissingDeltas;
    /**
     * Retrieves the missing deltas between the given sequence numbers
     */
    private fetchMissingDeltasCore;
    /**
     * Sorts pending ops and attempts to apply them
     */
    private processPendingOps;
    private updateLatestKnownOpSeqNumber;
}
//# sourceMappingURL=deltaManager.d.ts.map