/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { TypedEventEmitter, performanceNow } from "@fluid-internal/client-utils";
import { assert } from "@fluidframework/core-utils/internal";
import Deque from "double-ended-queue";
export class DeltaQueue extends TypedEventEmitter {
    get disposed() {
        return this.isDisposed;
    }
    /**
     * Whether or not the queue is paused.
     */
    get paused() {
        return this.pauseCount !== 0;
    }
    get length() {
        return this.q.length;
    }
    get idle() {
        return this.processingPromise === undefined && this.q.length === 0;
    }
    async waitTillProcessingDone() {
        return this.processingPromise ?? { count: 0, duration: 0 };
    }
    /**
     * @param worker - A callback to process a delta.
     */
    constructor(worker) {
        super();
        this.worker = worker;
        this.isDisposed = false;
        this.q = new Deque();
        /**
         * Tracks the number of pause requests for the queue.
         * The DeltaQueue is created initially paused.
         */
        this.pauseCount = 1;
    }
    dispose() {
        throw new Error("Not implemented.");
        this.isDisposed = true;
    }
    clear() {
        this.q.clear();
    }
    peek() {
        return this.q.peekFront();
    }
    toArray() {
        return this.q.toArray();
    }
    push(task) {
        try {
            this.q.push(task);
            this.emit("push", task);
            this.ensureProcessing();
        }
        catch (error) {
            this.emit("error", error);
        }
    }
    async pause() {
        this.pauseCount++;
        // If called from within the processing loop, we are in the middle of processing an op. Return a promise
        // that will resolve when processing has actually stopped.
        await this.waitTillProcessingDone();
    }
    resume() {
        assert(this.pauseCount > 0, 0x0f4 /* "Nonzero pause-count on resume()" */);
        this.pauseCount--;
        this.ensureProcessing();
    }
    /**
     * There are several actions that may need to kick off delta processing, so we want to guard against
     * accidental reentrancy. ensureProcessing can be called safely to start the processing loop if it is
     * not already started.
     */
    ensureProcessing() {
        if (this.anythingToProcess() && this.processingPromise === undefined) {
            // Use a resolved promise to start the processing on a separate stack.
            this.processingPromise = Promise.resolve()
                .then(() => {
                assert(this.processingPromise !== undefined, 0x37f /* reentrancy? */);
                const result = this.processDeltas();
                assert(this.processingPromise !== undefined, 0x380 /* reentrancy? */);
                // WARNING: Do not move next line to .finally() clause!
                // It runs async and creates a race condition where incoming ensureProcessing() call observes
                // from previous run while previous run is over (but finally clause was not scheduled yet)
                this.processingPromise = undefined;
                return result;
            })
                .catch((error) => {
                this.error = error;
                this.processingPromise = undefined;
                this.emit("error", error);
                return { count: 0, duration: 0 };
            });
            assert(this.processingPromise !== undefined, 0x381 /* processDeltas() should run async */);
        }
    }
    anythingToProcess() {
        return this.q.length > 0 && !this.paused && this.error === undefined;
    }
    /**
     * Executes the delta processing loop until a stop condition is reached.
     */
    processDeltas() {
        const start = performanceNow();
        let count = 0;
        // For grouping to work we must process all local messages immediately and in the single turn.
        // So loop over them until no messages to process, we have become paused, or hit an error.
        while (this.anythingToProcess()) {
            // Get the next message in the queue
            const next = this.q.shift();
            count++;
            // Process the message.
            // We know next is defined since we did a length check just prior to shifting.
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            this.worker(next);
            this.emit("op", next);
        }
        const duration = performanceNow() - start;
        if (this.q.length === 0) {
            this.emit("idle", count, duration);
        }
        return { count, duration };
    }
}
//# sourceMappingURL=deltaQueue.js.map