/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { type IContainer } from "@fluidframework/container-definitions/internal";
import { IRequest } from "@fluidframework/core-interfaces";
import { type ILoaderProps } from "./loader.js";
/**
 * Loads container and leaves it in a state where it does not process any ops.
 * Container instance returned by this function is in special mode where some functionality that is available in normal use will not work correctly
 * with instance of container returned by this function. Some examples:
 * 1. calling IContainer.connect() will have very little impact on this container as it will not process ops.
 * 2. functionality like waitContainerToCatchUp() or waiting for ops in any other way would hand infinitely, as this container is not processing ops
 * 3. No changes can be made to this container - they will be lost.
 *
 * If sequence number is provided, loads up to this sequence number and stops there, otherwise stops immediately after loading snapshot.
 * In all cases, container is returned disconnected & paused, or an exception is thrown
 * Notes:
 * 1. Ignores LoaderHeader.loadMode headers. Container is always returned with ops applied upt to provided sequence number,
 *    or no ops applied at all (if sequence number is not provided)
 * 2. This call can hang infinitately if disconnected from internet (or hit some other conditions, like 429 storm).
 *    Compare to Container.load() experience (with default settings) - it either returns failure right away, or succeeds, with
 *    ops fetching / delta connection happening in parallel / after container load flow, and thus providing an object (Container instance) to observe
 *    network connectivity issues / ability to cancel (IContainer.disconnect) or close container (IContainer.close)
 *    This flow needs to fetch ops (potentially connecting to delta connection), and any retriable errors on this path result in infinite retry.
 *    If you need to cancel that process, consider supplying AbortSignal parameter.
 * @param loaderProps - The loader props to use to load the container.
 * @param request - request identifying container instance / load parameters. LoaderHeader.loadMode headers are ignored (see above)
 * @param loadToSequenceNumber - optional sequence number. If provided, ops are processed up to this sequence number.
 * @param signal - optional abort signal that can be used to cancel waiting for the ops.
 * @returns IContainer instance
 *
 * @internal
 */
export declare function loadContainerPaused(loaderProps: ILoaderProps, request: IRequest, loadToSequenceNumber?: number, signal?: AbortSignal): Promise<IContainer>;
//# sourceMappingURL=loadPaused.d.ts.map