/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { IGetPendingLocalStateProps, IRuntime } from "@fluidframework/container-definitions/internal";
import type { IEventProvider, IEvent, ITelemetryBaseLogger } from "@fluidframework/core-interfaces";
import { IDocumentStorageService, IResolvedUrl, ISnapshot, ISnapshotTree, IVersion, ISequencedDocumentMessage } from "@fluidframework/driver-definitions/internal";
import { MonitoringContext } from "@fluidframework/telemetry-utils/internal";
import { ISerializableBlobContents } from "./containerStorageAdapter.js";
/**
 * This is very similar to {@link @fluidframework/protocol-definitions/internal#ISnapshot}, but the difference is
 * that the blobs of ISnapshot are of type ArrayBufferLike, while the blobs of this interface are serializable because
 * they are already converted to string.
 */
export interface SnapshotWithBlobs {
    /**
     * Snapshot from which container initially loaded.
     */
    baseSnapshot: ISnapshotTree;
    /**
     * Serializable blobs from the base snapshot. Used to load offline since
     * storage is not available.
     */
    snapshotBlobs: ISerializableBlobContents;
}
/**
 * State saved by a container at close time, to be used to load a new instance
 * of the container to the same state
 *
 * This is very similar to {@link @fluidframework/protocol-definitions/internal#ISnapshot}, but the difference is
 * that the blobs of ISnapshot are of type ArrayBufferLike, while the blobs of this interface are serializable because
 * they are already converted to string.
 *
 * @internal
 */
export interface IPendingContainerState extends SnapshotWithBlobs {
    /**
     * This container was attached (as opposed to IPendingDetachedContainerState.attached which is false)
     */
    attached: true;
    /**
     * Runtime-specific state that will be needed to properly rehydrate
     * (it's included in ContainerContext passed to instantiateRuntime)
     */
    pendingRuntimeState: unknown;
    /**
     * Any group snapshots (aka delay-loaded) we've downloaded from the service for this container
     */
    loadedGroupIdSnapshots?: Record<string, ISnapshotInfo>;
    /**
     * All ops since base snapshot sequence number up to the latest op
     * seen when the container was closed. Used to apply stashed (saved pending)
     * ops at the same sequence number at which they were made.
     */
    savedOps: ISequencedDocumentMessage[];
    /**
     * The Container's URL in the service, needed to hook up the driver during rehydration
     */
    url: string;
    /**
     * If the Container was connected when serialized, its clientId. Used as the initial clientId upon rehydration, until reconnected.
     */
    clientId?: string;
}
/**
 * State saved by a container in detached state, to be used to load a new instance
 * of the container to the same state (rehydrate)
 * @internal
 */
export interface IPendingDetachedContainerState extends SnapshotWithBlobs {
    /**
     * This container was not attached (as opposed to IPendingContainerState.attached which is true)
     */
    attached: false;
    /**
     * Indicates whether we expect the rehydrated container to have non-empty Detached Blob Storage
     */
    hasAttachmentBlobs: boolean;
    /**
     * Used by the memory blob storage to persisted attachment blobs
     */
    attachmentBlobs?: string;
    /**
     * Runtime-specific state that will be needed to properly rehydrate
     * (it's included in ContainerContext passed to instantiateRuntime)
     */
    pendingRuntimeState?: unknown;
}
export interface ISnapshotInfo extends SnapshotWithBlobs {
    snapshotSequenceNumber: number;
    snapshotFetchedTime?: number | undefined;
}
export type ISerializedStateManagerDocumentStorageService = Pick<IDocumentStorageService, "getSnapshot" | "getSnapshotTree" | "getVersions" | "readBlob"> & {
    loadedGroupIdSnapshots: Record<string, ISnapshot>;
};
interface ISerializerEvent extends IEvent {
    (event: "saved", listener: (dirty: boolean) => void): void;
}
/**
 * Helper class to manage the state of the container needed for proper serialization.
 *
 * It holds the pendingLocalState the container was rehydrated from (if any),
 * as well as the snapshot to be used for serialization.
 * It also keeps track of container dirty state and which local ops have been processed
 */
export declare class SerializedStateManager {
    private readonly pendingLocalState;
    private readonly storageAdapter;
    private readonly _offlineLoadEnabled;
    private readonly containerDirty;
    private readonly supportGetSnapshotApi;
    private readonly processedOps;
    private readonly mc;
    private snapshot;
    private latestSnapshot;
    private _refreshSnapshotP;
    private readonly lastSavedOpSequenceNumber;
    private readonly refreshTimer;
    private readonly snapshotRefreshTimeoutMs;
    /**
     * @param pendingLocalState - The pendingLocalState being rehydrated, if any (undefined when loading directly from storage)
     * @param subLogger - Container's logger to use as parent for our logger
     * @param storageAdapter - Storage adapter for fetching snapshots
     * @param _offlineLoadEnabled - Is serializing/rehydrating containers allowed?
     * @param containerEvent - Source of the "saved" event when the container has all its pending state uploaded
     * @param containerDirty - Is the container "dirty"? That's the opposite of "saved" - there is pending state that may not have been received yet by the service.
     */
    constructor(pendingLocalState: IPendingContainerState | undefined, subLogger: ITelemetryBaseLogger, storageAdapter: ISerializedStateManagerDocumentStorageService, _offlineLoadEnabled: boolean, containerEvent: IEventProvider<ISerializerEvent>, containerDirty: () => boolean, supportGetSnapshotApi: () => boolean, snapshotRefreshTimeoutMs?: number);
    get offlineLoadEnabled(): boolean;
    /**
     * Promise that will resolve (or reject) once we've tried to download the latest snapshot(s) from storage
     * only intended to be used for testing purposes.
     * @returns The snapshot sequence number associated with the latest fetched snapshot
     */
    get refreshSnapshotP(): Promise<number> | undefined;
    /**
     * Called whenever an incoming op is processed by the Container
     */
    addProcessedOp(message: ISequencedDocumentMessage): void;
    /**
     * This wraps the basic functionality of fetching the snapshot for this container during Container load.
     *
     * If we have pendingLocalState, we get the snapshot from there.
     * Otherwise, fetch it from storage (according to specifiedVersion if provided).
     *
     * @param specifiedVersion - If a version is specified and we don't have pendingLocalState, fetch this version from storage.
     * @param supportGetSnapshotApi - a boolean indicating whether to use the fetchISnapshot or fetchISnapshotTree.
     * @returns The snapshot to boot the container from
     */
    fetchSnapshot(specifiedVersion: string | undefined): Promise<{
        baseSnapshot: ISnapshot | ISnapshotTree;
        version: IVersion | undefined;
    }>;
    private tryRefreshSnapshot;
    /**
     * Fetch the latest snapshot for the container, including delay-loaded groupIds if pendingLocalState was provided and contained any groupIds.
     * Note that this will update the StorageAdapter's cached snapshots for the groupIds (if present)
     *
     * @param supportGetSnapshotApi - a boolean indicating whether to use the fetchISnapshot or fetchISnapshotTree (must be true to fetch by groupIds)
     */
    private refreshLatestSnapshot;
    /**
     * Updates class snapshot and processedOps if we have a new snapshot and it's among processedOps range.
     */
    private updateSnapshotAndProcessedOpsMaybe;
    /**
     * When the Container attaches, we need to stash the initial snapshot (a form of the attach summary).
     * This method is only meant to be used by Container.attach() to set the initial
     * base snapshot when attaching.
     * @param snapshot - snapshot and blobs collected while attaching (a form of the attach summary)
     */
    setInitialSnapshot(snapshot: SnapshotWithBlobs | undefined): void;
    /**
     * Assembles and serializes the {@link IPendingContainerState} for the container,
     * to be stored and used to rehydrate the container at a later time.
     */
    getPendingLocalState(props: IGetPendingLocalStateProps, clientId: string | undefined, runtime: Pick<IRuntime, "getPendingLocalState">, resolvedUrl: IResolvedUrl): Promise<string>;
}
/**
 * Retrieves the most recent snapshot and returns its info.
 *
 * @param mc - The monitoring context.
 * @param storageAdapter - The storage adapter providing methods to retrieve the snapshot.
 * @param supportGetSnapshotApi - a boolean indicating whether to use the fetchISnapshot or fetchISnapshotTree.
 * @returns a SnapshotInfo object containing the snapshot tree, snapshot blobs and its sequence number.
 */
export declare function getLatestSnapshotInfo(mc: MonitoringContext, storageAdapter: ISerializedStateManagerDocumentStorageService, supportGetSnapshotApi: boolean): Promise<ISnapshotInfo | undefined>;
/**
 * Fetches an ISnapshot from a storage adapter based on the specified version.
 *
 * @param mc - The monitoring context.
 * @param storageAdapter - The storage adapter providing a getSnapshot method to retrieve the ISnapshot and version.
 * @param specifiedVersion - An optional version string specifying the version of the snapshot tree to fetch.
 * @returns - The fetched snapshot tree and its version.
 */
export declare function fetchISnapshot(mc: MonitoringContext, storageAdapter: Pick<IDocumentStorageService, "getSnapshot">, specifiedVersion: string | undefined): Promise<{
    snapshot?: ISnapshot;
    version?: IVersion;
}>;
/**
 * Fetches an ISnapshotTree from a storage adapter based on the specified version.
 *
 * @param mc - The monitoring context.
 * @param storageAdapter - The storage adapter providing methods to retrieve the ISnapshotTree and version.
 * @param specifiedVersion - An optional version string specifying the version of the snapshot tree to fetch.
 * @returns - The fetched snapshot tree and its version.
 */
export declare function fetchISnapshotTree(mc: MonitoringContext, storageAdapter: Pick<IDocumentStorageService, "getSnapshotTree" | "getVersions">, specifiedVersion: string | undefined): Promise<{
    snapshot?: ISnapshotTree;
    version?: IVersion | undefined;
}>;
export {};
//# sourceMappingURL=serializedStateManager.d.ts.map